@@ -134,10 +134,6 @@ struct wrap_driver *load_wrap_driver(struct wrap_device *wd)
 		}
 		DBGTRACE1("driver %s is loaded", wrap_driver->name);
 	}
-
-	nt_spin_lock(&loader_lock);
-	InsertTailList(&wrap_driver->wrap_devices, &wd->list);
-	nt_spin_unlock(&loader_lock);
 	TRACEEXIT1(return wrap_driver);
 }
 
@@ -247,7 +243,7 @@ struct wrap_bin_file *get_bin_file(char *bin_file_name)
 		DBGTRACE1("coudln't find bin file '%s'", bin_file_name);
 		return NULL;
 	}
-	
+
 	if (!driver->bin_files[i].data) {
 		char *argv[] = {"loadndisdriver", WRAP_CMD_LOAD_BIN_FILE,
 #if defined DEBUG && DEBUG >= 1
@@ -441,18 +437,16 @@ static void unload_wrap_device(struct wrap_device *wd)
 		setting = container_of(cur, struct wrap_device_setting, list);
 		kfree(setting);
 	}
-	InitializeListHead(&wd->list);
 	TRACEEXIT1(return);
 }
 
+/* this function is called while holding load_lock spinlock */
 void unload_wrap_driver(struct wrap_driver *driver)
 {
 	int i;
 	struct driver_object *drv_obj;
-	struct nt_list *cur, *next;
 
 	TRACEENTER1("unloading driver: %s (%p)", driver->name, driver);
-	RemoveEntryList(&driver->list);
 	DBGTRACE1("freeing %d images", driver->num_pe_images);
 	drv_obj = driver->drv_obj;
 	for (i = 0; i < driver->num_pe_images; i++)
@@ -468,21 +462,15 @@ void unload_wrap_driver(struct wrap_driver *driver)
 		if (driver->bin_files[i].data)
 			vfree(driver->bin_files[i].data);
 	}
-	DBGTRACE2("");
 	if (driver->bin_files)
 		kfree(driver->bin_files);
-	DBGTRACE2("");
 	RtlFreeUnicodeString(&drv_obj->name);
-	nt_list_for_each_safe(cur, next, &driver->wrap_devices) {
-		struct wrap_device *wd;
-		wd = container_of(cur, struct wrap_device, list);
-		RemoveEntryList(&wd->list);
-	}
+	RemoveEntryList(&driver->list);
 	/* this frees driver */
 	free_custom_extensions(drv_obj->drv_ext);
 	kfree(drv_obj->drv_ext);
 	DBGTRACE1("drv_obj: %p", drv_obj);
-		
+
 	TRACEEXIT1(return);
 }
 
@@ -492,8 +480,7 @@ static int start_wrap_driver(struct wrap_driver *driver)
 	int i;
 	NTSTATUS ret, res;
 	struct driver_object *drv_obj;
-	UINT (*entry)(struct driver_object *obj,
-		      struct unicode_string *path) STDCALL;
+	typeof(driver->pe_images[0].entry) entry;
 
 	TRACEENTER1("%s", driver->name);
 	drv_obj = driver->drv_obj;
@@ -686,7 +673,6 @@ static int register_devices(struct load_devices *load_devices)
 		struct wrap_device *wd;
 
 		wd = &wrap_devices[num_pci + num_usb];
-		InitializeListHead(&wd->list);
 		InitializeListHead(&wd->settings);
 		strncpy(wd->driver_name, device->driver_name,
 			sizeof(wd->driver_name));
