@@ -25,16 +25,33 @@
 extern struct nt_list wrap_drivers;
 extern NT_SPIN_LOCK ntoskernel_lock, loader_lock;
 
+static struct workqueue_struct *ndis_wq;
+static void ndis_worker(void *dummy);
+static struct work_struct ndis_work;
+static struct nt_list ndis_worker_list;
+static NT_SPIN_LOCK ndis_work_list_lock;
+
 /* ndis_init is called once when module is loaded */
 int ndis_init(void)
 {
+#ifdef USE_OWN_WORKQUEUE
+	ndis_wq = create_singlethread_workqueue("ndis_wq");
+#else
+	ndis_wq = NULL;
+#endif
+	InitializeListHead(&ndis_worker_list);
+	nt_spin_lock_init(&ndis_work_list_lock);
+	INIT_WORK(&ndis_work, ndis_worker, NULL);
+
 	return 0;
 }
 
 /* ndis_exit is called once when module is removed */
 void ndis_exit(void)
 {
-	/* TODO: free all packets in all pools */
+#ifdef USE_OWN_WORKQUEUE
+	destroy_workqueue(ndis_wq);
+#endif
 	TRACEEXIT1(return);
 }
 
@@ -276,8 +293,8 @@ STDCALL void WRAP_EXPORT(NdisOpenConfigurationKeyByIndex)
 	 struct unicode_string *key, void **subkeyhandle)
 {
 	TRACEENTER2("index: %u", index);
-	*subkeyhandle = handle;
-	*status = NDIS_STATUS_SUCCESS;
+//	*subkeyhandle = handle;
+	*status = NDIS_STATUS_FAILURE;
 	TRACEEXIT2(return);
 }
 
@@ -361,25 +378,18 @@ STDCALL void WRAP_EXPORT(NdisGetSystemUpTime)
 	TRACEEXIT5(return);
 }
 
-/* called as macro */
 STDCALL ULONG WRAP_EXPORT(NDIS_BUFFER_TO_SPAN_PAGES)
 	(ndis_buffer *buffer)
 {
-	ULONG_PTR start;
-	ULONG n;
-
-	TRACEENTER4("");
+	ULONG n, length;
 
 	if (buffer == NULL)
 		return 0;
-
 	if (MmGetMdlByteCount(buffer) == 0)
 		return 1;
-
-	start = (ULONG_PTR)(MmGetMdlVirtualAddress(buffer));
-	n = SPAN_PAGES(start, MmGetMdlByteCount(buffer));
-	DBGTRACE4("pages = %u", n);
-	TRACEEXIT4(return n);
+	length = MmGetMdlByteCount(buffer);
+	n = SPAN_PAGES(MmGetMdlVirtualAddress(buffer), length);
+	TRACEEXIT3(return n);
 }
 
 STDCALL void WRAP_EXPORT(NdisGetBufferPhysicalArraySize)
@@ -570,6 +580,7 @@ STDCALL void WRAP_EXPORT(NdisWriteConfiguration)
 		}
 	} else
 		*status = NDIS_STATUS_RESOURCES;
+
 	RtlFreeAnsiString(&ansi);
 	TRACEEXIT2(return);
 }
@@ -609,19 +620,29 @@ STDCALL void WRAP_EXPORT(NdisInitUnicodeString)
 STDCALL NDIS_STATUS WRAP_EXPORT(NdisAnsiStringToUnicodeString)
 	(struct unicode_string *dst, struct ansi_string *src)
 {
+	BOOLEAN alloc;
 	TRACEENTER2("");
 	if (dst == NULL || src == NULL)
 		TRACEEXIT2(return NDIS_STATUS_FAILURE);
-	TRACEEXIT2(return RtlAnsiStringToUnicodeString(dst, src, FALSE));
+	if (dst->buf == NULL)
+		alloc = TRUE;
+	else
+		alloc = FALSE;
+	TRACEEXIT2(return RtlAnsiStringToUnicodeString(dst, src, alloc));
 }
 
 STDCALL NDIS_STATUS WRAP_EXPORT(NdisUnicodeStringToAnsiString)
 	(struct ansi_string *dst, struct unicode_string *src)
 {
+	BOOLEAN alloc;
 	TRACEENTER2("");
 	if (dst == NULL || src == NULL)
 		TRACEEXIT2(return NDIS_STATUS_FAILURE);
-	TRACEEXIT2(return RtlUnicodeStringToAnsiString(dst, src, FALSE));
+	if (dst->buf == NULL)
+		alloc = TRUE;
+	else
+		alloc = FALSE;
+	TRACEEXIT2(return RtlUnicodeStringToAnsiString(dst, src, alloc));
 }
 
 STDCALL void WRAP_EXPORT(NdisMSetAttributesEx)
@@ -728,8 +749,11 @@ STDCALL void WRAP_EXPORT(NdisMQueryAdapterResources)
 	resource_length = sizeof(struct cm_partial_resource_list) +
 		sizeof(struct cm_partial_resource_descriptor) *
 		(list->count - 1);
-	DBGTRACE2("wnd: %p. buf: %p, len: %d (%d)", wnd,
-		  resource_list, *size, resource_length);
+	DBGTRACE2("wnd: %p. buf: %p, len: %d (%d), %p %d %d", wnd,
+		  resource_list, *size, resource_length,
+		  &list->partial_descriptors[list->count-1],
+		  list->partial_descriptors[list->count-1].u.interrupt.level,
+		  list->partial_descriptors[list->count-1].u.interrupt.vector);
 	if (*size == 0) {
 		*size = resource_length;
 		*status = NDIS_STATUS_BUFFER_TOO_SHORT;
@@ -748,7 +772,7 @@ STDCALL NDIS_STATUS WRAP_EXPORT(NdisMPciAssignResources)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
 
-	TRACEENTER2("%p", wnd);
+	TRACEENTER2("%p, %p", wnd, wnd->wd->resource_list);
 	*resources = &wnd->wd->resource_list->list->partial_resource_list;
 	TRACEEXIT2(return NDIS_STATUS_SUCCESS);
 }
@@ -784,8 +808,7 @@ STDCALL void WRAP_EXPORT(NdisMUnmapIoSpace)
 STDCALL void WRAP_EXPORT(NdisAllocateSpinLock)
 	(struct ndis_spinlock *lock)
 {
-	TRACEENTER4("lock %p", lock);
-
+	DBGTRACE4("lock %p, %lu", lock, lock->klock);
 	KeInitializeSpinLock(&lock->klock);
 	lock->irql = PASSIVE_LEVEL;
 	TRACEEXIT4(return);
@@ -794,14 +817,14 @@ STDCALL void WRAP_EXPORT(NdisAllocateSpinLock)
 STDCALL void WRAP_EXPORT(NdisFreeSpinLock)
 	(struct ndis_spinlock *lock)
 {
-	TRACEENTER4("lock %p", lock);
+	DBGTRACE4("lock %p, %lu", lock, lock->klock);
 	TRACEEXIT4(return);
 }
 
 STDCALL void WRAP_EXPORT(NdisAcquireSpinLock)
 	(struct ndis_spinlock *lock)
 {
-	TRACEENTER6("lock %p", lock);
+	DBGTRACE6("lock %p, %lu", lock, lock->klock);
 	lock->irql = nt_spin_lock_irql(&lock->klock, DISPATCH_LEVEL);
 	TRACEEXIT6(return);
 }
@@ -809,7 +832,7 @@ STDCALL void WRAP_EXPORT(NdisAcquireSpinLock)
 STDCALL void WRAP_EXPORT(NdisReleaseSpinLock)
 	(struct ndis_spinlock *lock)
 {
-	TRACEENTER6("lock %p", lock);
+	DBGTRACE6("lock %p, %lu", lock, lock->klock);
 	nt_spin_unlock_irql(&lock->klock, lock->irql);
 	TRACEEXIT6(return);
 }
@@ -844,7 +867,10 @@ STDCALL NDIS_STATUS WRAP_EXPORT(NdisMAllocateMapRegisters)
 	 NDIS_DMA_SIZE dmasize, ULONG basemap, ULONG max_buf_size)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
-	TRACEENTER2("%d %d %d %d", dmachan, dmasize, basemap, size);
+	TRACEENTER2("%d %d %d %d", dmachan, dmasize, basemap, max_buf_size);
+
+	if (max_buf_size > PAGE_SIZE)
+	    WARNING("buffer size too big: %d", max_buf_size);
 
 //	if (basemap > 64)
 //		return NDIS_STATUS_RESOURCES;
@@ -886,20 +912,46 @@ STDCALL void WRAP_EXPORT(NdisMAllocateSharedMemory)
 	(struct ndis_miniport_block *nmb, ULONG size,
 	 BOOLEAN cached, void **virt, NDIS_PHY_ADDRESS *phys)
 {
-	dma_addr_t p;
+	dma_addr_t dma_addr;
 	struct wrap_device *wd = nmb->wnd->wd;
 
 	TRACEENTER3("map count: %d, size: %u, cached: %d",
 		    nmb->wnd->dma_map_count, size, cached);
 
-	*virt = PCI_DMA_ALLOC_COHERENT(wd->pci.pdev, size, &p);
+	*virt = PCI_DMA_ALLOC_COHERENT(wd->pci.pdev, size, &dma_addr);
 	if (!*virt) {
-		ERROR("failed to allocate DMA coherent memory; "
-		      "Windows driver requested %d bytes of "
-		      "%scached memory", size, cached ? "" : "un-");
+		ERROR("couldn't allocate %d bytes of %scached DMA memory",
+		      size, cached ? "" : "un-");
 	}
-	*phys = p;
-	DBGTRACE3("allocated shared memory: %p", *virt);
+	*phys = dma_addr;
+	TRACEEXIT3(return);
+}
+
+static void ndis_worker(void *dummy)
+{
+	KIRQL irql;
+	struct ndis_work_entry *ndis_work_entry;
+	struct nt_list *ent;
+	struct ndis_work_item *ndis_work_item;
+
+	WORKENTER("");
+	while (1) {
+		irql = nt_spin_lock_irql(&ndis_work_list_lock, DISPATCH_LEVEL);
+		ent = RemoveHeadList(&ndis_worker_list);
+		nt_spin_unlock_irql(&ndis_work_list_lock, irql);
+		if (!ent)
+			break;
+		ndis_work_entry = container_of(ent, struct ndis_work_entry,
+					       list);
+		ndis_work_item = ndis_work_entry->ndis_work_item;
+		WORKTRACE("%p: calling work at %p with %p", ndis_work_item,
+			  ndis_work_item->func, ndis_work_item->ctx);
+		LIN2WIN2(ndis_work_item->func, ndis_work_item,
+			 ndis_work_item->ctx);
+		WORKTRACE("%p done", ndis_work_item);
+		kfree(ndis_work_entry);
+	}
+	WORKEXIT(return);
 }
 
 STDCALL void alloc_shared_memory_async(void *arg1, void *arg2)
@@ -938,7 +990,7 @@ STDCALL NDIS_STATUS WRAP_EXPORT(NdisMAllocateSharedMemoryAsync)
 	alloc_shared_mem->size = size;
 	alloc_shared_mem->cached = cached;
 	alloc_shared_mem->ctx = ctx;
-	if (schedule_wrap_work_item(alloc_shared_memory_async,
+	if (schedule_ntos_work_item(alloc_shared_memory_async,
 				    wnd, alloc_shared_mem, FALSE))
 		TRACEEXIT3(return NDIS_STATUS_FAILURE);
 	TRACEEXIT3(return NDIS_STATUS_PENDING);
@@ -950,7 +1002,6 @@ STDCALL void WRAP_EXPORT(NdisMFreeSharedMemory)
 {
 	struct wrap_device *wd = nmb->wnd->wd;
 	TRACEENTER3("");
-	/* FIXME: do USB drivers call this? */
 	PCI_DMA_FREE_COHERENT(wd->pci.pdev, size, virt, addr);
 	TRACEEXIT3(return);
 }
@@ -1049,6 +1100,10 @@ STDCALL void WRAP_EXPORT(NdisFreeBuffer)
 	KIRQL irql;
 
 	TRACEENTER4("descr: %p", descr);
+	if (!descr) {
+		ERROR("buffer is NULL");
+		TRACEEXIT4(return);
+	}
 	pool = descr->process;
 	if (!pool) {
 		ERROR("pool for descriptor %p is invalid", descr);
@@ -1074,7 +1129,7 @@ STDCALL void WRAP_EXPORT(NdisFreeBuffer)
 STDCALL void WRAP_EXPORT(NdisFreeBufferPool)
 	(struct ndis_buffer_pool *pool)
 {
-	ndis_buffer *cur, *prev;
+	ndis_buffer *cur, *next;
 	KIRQL irql;
 
 	DBGTRACE3("pool: %p", pool);
@@ -1085,10 +1140,10 @@ STDCALL void WRAP_EXPORT(NdisFreeBufferPool)
 	irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
 	cur = pool->free_descr;
 	while (cur) {
-		prev = cur;
-		cur = cur->next;
-		prev->process = NULL;
-		free_mdl(prev);
+		next = cur->next;
+		cur->process = NULL;
+		free_mdl(cur);
+		cur = next;
 	}
 	nt_spin_unlock_irql(&pool->lock, irql);
 	kfree(pool);
@@ -1110,7 +1165,7 @@ STDCALL void WRAP_EXPORT(NdisQueryBuffer)
 	if (virt)
 		*virt = MmGetSystemAddressForMdl(buffer);
 	*length = MmGetMdlByteCount(buffer);
-	DBGTRACE4("%p, %u", *virt, *length);
+	DBGTRACE4("%u", *length);
 	TRACEEXIT4(return);
 }
 
@@ -1122,7 +1177,7 @@ STDCALL void WRAP_EXPORT(NdisQueryBufferSafe)
 	if (virt)
 		*virt = MmGetSystemAddressForMdlSafe(buffer, priority);
 	*length = MmGetMdlByteCount(buffer);
-	DBGTRACE4("%p, %u", *virt, *length);
+	DBGTRACE4("%u", *length);
 }
 
 STDCALL void *WRAP_EXPORT(NdisBufferVirtualAddress)
@@ -1579,15 +1634,16 @@ STDCALL void WRAP_EXPORT(NdisSend)
 			*status = oob_data->status;
 			switch (*status) {
 			case NDIS_STATUS_SUCCESS:
-				sendpacket_done(wnd, packet);
+				free_tx_packet(wnd, packet, *status);
 				break;
 			case NDIS_STATUS_PENDING:
 				break;
 			case NDIS_STATUS_RESOURCES:
-				wnd->send_ok = 0;
+				wnd->tx_ok = 0;
 				break;
 			case NDIS_STATUS_FAILURE:
 			default:
+				free_tx_packet(wnd, packet, *status);
 				break;
 			}
 		} else {
@@ -1600,14 +1656,16 @@ STDCALL void WRAP_EXPORT(NdisSend)
 		lower_irql(irql);
 		switch (*status) {
 		case NDIS_STATUS_SUCCESS:
-			sendpacket_done(wnd, packet);
+			free_tx_packet(wnd, packet, *status);
 			break;
 		case NDIS_STATUS_PENDING:
 			break;
 		case NDIS_STATUS_RESOURCES:
-			wnd->send_ok = 0;
+			wnd->tx_ok = 0;
 			break;
 		case NDIS_STATUS_FAILURE:
+		default:
+			free_tx_packet(wnd, packet, *status);
 			break;
 		}
 	}
@@ -1722,7 +1780,7 @@ STDCALL void WRAP_EXPORT(NdisMRegisterAdapterShutdownHandler)
 	(struct ndis_miniport_block *nmb, void *ctx, void *func)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
-	TRACEENTER1("sp:%p", get_sp());
+	TRACEENTER1("%p", func);
 	wnd->wd->driver->ndis_driver->miniport.shutdown = func;
 	wnd->shutdown_ctx = ctx;
 }
@@ -1731,7 +1789,6 @@ STDCALL void WRAP_EXPORT(NdisMDeregisterAdapterShutdownHandler)
 	(struct ndis_miniport_block *nmb)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
-	TRACEENTER1("sp:%p", get_sp());
 	wnd->wd->driver->ndis_driver->miniport.shutdown = NULL;
 	wnd->shutdown_ctx = NULL;
 }
@@ -1787,9 +1844,8 @@ STDCALL NDIS_STATUS WRAP_EXPORT(NdisMRegisterInterrupt)
 	 BOOLEAN shared, enum kinterrupt_mode mode)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
-	TRACEENTER1("%p, vector:%d, level:%d, req_isr:%d, shared:%d, "
-		    "mode:%d sp:%p", ndis_irq, vector, level, req_isr,
-		    shared, mode, get_sp());
+	TRACEENTER1("%p, vector:%d, level:%d, req_isr:%d, shared:%d, mode:%d",
+		    ndis_irq, vector, level, req_isr, shared, mode);
 
 	ndis_irq->irq.irq = vector;
 	ndis_irq->wnd = wnd;
@@ -1846,6 +1902,7 @@ STDCALL BOOLEAN WRAP_EXPORT(NdisMSynchronizeWithInterrupt)
 	nt_spin_lock_irqsave(&ndis_irq->lock, flags);
 	ret = LIN2WIN1(sync_func, ctx);
 	nt_spin_unlock_irqrestore(&ndis_irq->lock, flags);
+	DBGTRACE6("ret: %d", ret);
 	TRACEEXIT6(return ret);
 }
 
@@ -1861,21 +1918,21 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatus)
 
 	TRACEENTER2("status=0x%x len=%d", status, len);
 	switch (status) {
-	case  NDIS_STATUS_MEDIA_DISCONNECT:
+	case NDIS_STATUS_MEDIA_DISCONNECT:
 		if (wnd->link_status != 0) {
 			wnd->link_status = 0;
+			wnd->tx_ok = 0;
 			set_bit(LINK_STATUS_CHANGED, &wnd->wrap_ndis_work);
+			schedule_wrap_work(&wnd->wrap_ndis_worker);
 		}
-		wnd->link_status = 0;
-		wnd->send_ok = 0;
 		break;
 	case NDIS_STATUS_MEDIA_CONNECT:
 		if (wnd->link_status != 1) {
 			wnd->link_status = 1;
+			wnd->tx_ok = 1;
 			set_bit(LINK_STATUS_CHANGED, &wnd->wrap_ndis_work);
+			schedule_wrap_work(&wnd->wrap_ndis_worker);
 		}
-		wnd->link_status = 1;
-		wnd->send_ok = 1;
 		break;
 	case NDIS_STATUS_MEDIA_SPECIFIC_INDICATION:
 		if (!buf)
@@ -1976,11 +2033,11 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatus)
 		}
 		break;
 	default:
-		DBGTRACE4("unknown status: %08X", status);
+		DBGTRACE2("unknown status: %08X", status);
 		break;
 	}
 
-	TRACEEXIT1(return);
+	TRACEEXIT2(return);
 }
 
 /* called via function pointer */
@@ -1989,9 +2046,9 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatusComplete)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER2("%p", wnd);
-	schedule_work(&wnd->wrap_ndis_worker);
-	if (wnd->send_ok)
-		schedule_work(&wnd->xmit_work);
+	schedule_wrap_work(&wnd->wrap_ndis_worker);
+	if (wnd->tx_ok)
+		schedule_wrap_work(&wnd->tx_work);
 }
 
 STDCALL void return_packet(void *arg1, void *arg2)
@@ -2017,7 +2074,7 @@ STDCALL void
 NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 			   struct ndis_packet **packets, UINT nr_packets)
 {
-	struct wrap_ndis_device *wnd = nmb->wnd;
+	struct wrap_ndis_device *wnd;
 	ndis_buffer *buffer;
 	struct ndis_packet *packet;
 	struct sk_buff *skb;
@@ -2025,7 +2082,8 @@ NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 	struct ndis_packet_oob_data *oob_data;
 	void *virt;
 
-	TRACEENTER3("%d", nr_packets);
+	TRACEENTER3("%p, %d", nmb, nr_packets);
+	wnd = nmb->wnd;
 	for (i = 0; i < nr_packets; i++) {
 		packet = packets[i];
 		if (!packet) {
@@ -2078,7 +2136,7 @@ NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 		 * MiniportReturnPacket from here is not correct - the
 		 * driver doesn't expect it (at least Centrino driver
 		 * crashes) */
-		schedule_wrap_work_item(return_packet, wnd, packet, FALSE);
+		schedule_ntos_work_item(return_packet, wnd, packet, FALSE);
 	}
 	TRACEEXIT3(return);
 }
@@ -2099,13 +2157,14 @@ NdisMSendComplete(struct ndis_miniport_block *nmb, struct ndis_packet *packet,
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER3("%p, %08x", packet, status);
-	sendpacket_done(wnd, packet);
+	free_tx_packet(wnd, packet, status);
 	/* In case a serialized driver has requested a pause by returning
 	 * NDIS_STATUS_RESOURCES we need to give the send-code a kick again.
 	 */
-	if (wnd->send_ok == 0) {
-		wnd->send_ok = 1;
-		schedule_work(&wnd->xmit_work);
+	if (wnd->tx_ok == 0) {
+		wnd->tx_ok = 1;
+		DBGTRACE3("%d, %d", wnd->tx_ring_start, wnd->tx_ring_end);
+		schedule_wrap_work(&wnd->tx_work);
 	}
 	TRACEEXIT3(return);
 }
@@ -2125,11 +2184,9 @@ NdisMSendResourcesAvailable(struct ndis_miniport_block *nmb)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER3("");
-	/* sending packets immediately seem to result in NDIS_STATUS_FAILURE,
-	   so wait for a while before sending the packet again */
-	mdelay(5);
-	wnd->send_ok = 1;
-	schedule_work(&wnd->xmit_work);
+	DBGTRACE3("%d, %d", wnd->tx_ring_start, wnd->tx_ring_end);
+	wnd->tx_ok = 1;
+	schedule_wrap_work(&wnd->tx_work);
 	TRACEEXIT3(return);
 }
 
@@ -2163,7 +2220,7 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 		unsigned int bytes_txed;
 		NDIS_STATUS res;
 
-		NdisAllocatePacket(&res, &packet, wnd->wrapper_packet_pool);
+		NdisAllocatePacket(&res, &packet, wnd->tx_packet_pool);
 		if (res != NDIS_STATUS_SUCCESS) {
 			wnd->stats.rx_dropped++;
 			TRACEEXIT3(return);
@@ -2184,7 +2241,7 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 			if (skb) {
 				memcpy(skb_put(skb, header_size), header,
 				       header_size);
-				memcpy(skb_put(skb, look_ahead_size), 
+				memcpy(skb_put(skb, look_ahead_size),
 				       look_ahead, look_ahead_size);
 				buffer = packet->private.buffer_head;
 				while (buffer) {
@@ -2353,7 +2410,7 @@ STDCALL void WRAP_EXPORT(NdisMSleep)
 	unsigned long delay;
 
 	TRACEENTER4("%p: us: %u", current, us);
-	delay = USEC_TO_HZ(us) + 1;
+	delay = USEC_TO_HZ(us);
 	set_current_state(TASK_INTERRUPTIBLE);
 	schedule_timeout(delay);
 	DBGTRACE4("%p: done", current);
@@ -2416,7 +2473,7 @@ STDCALL struct nt_list *WRAP_EXPORT(NdisInterlockedRemoveHeadList)
 }
 
 STDCALL NDIS_STATUS WRAP_EXPORT(NdisMInitializeScatterGatherDma)
-	(struct ndis_miniport_block *nmb, UCHAR dma_size, ULONG max_phy_map)
+	(struct ndis_miniport_block *nmb, BOOLEAN dma_size, ULONG max_phy_map)
 {
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER2("dma_size=%d, maxtransfer=%u", dma_size, max_phy_map);
@@ -2499,14 +2556,26 @@ NdisMResetComplete(struct ndis_miniport_block *nmb, NDIS_STATUS status,
 }
 
 STDCALL NDIS_STATUS WRAP_EXPORT(NdisScheduleWorkItem)
-	(struct ndis_sched_work_item *ndis_sched_work_item)
+	(struct ndis_work_item *ndis_work_item)
 {
-	TRACEENTER3("%p", ndis_sched_work_item);
-	schedule_wrap_work_item(ndis_sched_work_item->func,
-				ndis_sched_work_item,
-				ndis_sched_work_item->ctx, TRUE);
+	struct ndis_work_entry *ndis_work_entry;
+	KIRQL irql;
 
+	TRACEENTER3("%p", ndis_work_item);
+	ndis_work_entry = kmalloc(sizeof(*ndis_work_entry), GFP_ATOMIC);
+	if (!ndis_work_entry)
+		BUG();
+
+	ndis_work_entry->ndis_work_item = ndis_work_item;
+
+	irql = nt_spin_lock_irql(&ndis_work_list_lock, DISPATCH_LEVEL);
+	InsertTailList(&ndis_worker_list, &ndis_work_entry->list);
+	nt_spin_unlock_irql(&ndis_work_list_lock, irql);
+
+	WORKTRACE("scheduling %p", ndis_work_item);
+	schedule_ndis_work(&ndis_work);
 	TRACEEXIT3(return NDIS_STATUS_SUCCESS);
+
 }
 
 STDCALL void WRAP_EXPORT(NdisMGetDeviceProperty)
@@ -2594,5 +2663,20 @@ void init_nmb_functions(struct ndis_miniport_block *nmb)
 	nmb->eth_rx_indicate = WRAP_FUNC_PTR(EthRxIndicateHandler);
 	nmb->eth_rx_complete = WRAP_FUNC_PTR(EthRxComplete);
 	nmb->td_complete = WRAP_FUNC_PTR(NdisMTransferDataComplete);
+
+	/* TODO: setup pointers to
+
+	   NdisMWanSendComplete
+	   NdisMWanIndicateReceive
+	   NdisMWanIndicateReceiveComplete
+	   
+	   NdisMTrIndicateReceive
+	   NdisMTrIndicateReceiveComplete
+
+	   NdisMFddiIndicateReceive
+	   NdisMFddiIndicateReceiveComplete
+	   NdisMArcIndicateReceive
+	   NdisMArcIndicateReceiveComplete
+	 */
 }
 
