@@ -34,7 +34,8 @@ typedef LONG ndis_rssi;
 typedef ULONG ndis_key_index;
 typedef ULONG ndis_tx_power_level;
 typedef ULONGULONG ndis_key_rsc;
-typedef UCHAR ndis_rates[NDIS_MAX_RATES_EX];
+typedef UCHAR ndis_rates[NDIS_MAX_RATES];
+typedef UCHAR ndis_rates_ex[NDIS_MAX_RATES_EX];
 typedef UCHAR mac_address[ETH_ALEN];
 typedef ULONG ndis_fragmentation_threshold;
 typedef ULONG ndis_rts_threshold;
@@ -336,25 +337,14 @@ struct ndis_rw_lock {
 	union ndis_rw_lock_refcount ref_count[MAXIMUM_PROCESSORS];
 };
 
-struct ndis_sched_work_item {
-	void *ctx;
-	WRAP_WORK_FUNC func;
-	UCHAR reserved[8 * sizeof(void *)];
-};
-
-struct ndis_alloc_mem_work_item {
-	unsigned long size;
-	char cached;
-	void *ctx;
-};
-
-struct ndis_free_mem_work_item {
-	void *addr;
-};
+struct ndis_work_item;
+typedef void (*NDIS_PROC)(struct ndis_work_item *, void *) STDCALL;
 
 struct ndis_work_item {
-	void *context;
-	void *routine;
+	void *ctx;
+	/* this should be NDIS_PROC, but we masquerade it as
+	 * NTOS_WORK_FUNC so we can use wrap_worker */
+	NTOS_WORK_FUNC func;
 	UCHAR reserved[8 * sizeof(void *)];
 };
 
@@ -364,6 +354,11 @@ struct alloc_shared_mem {
 	BOOLEAN cached;
 };
 
+struct ndis_work_entry {
+	struct nt_list list;
+	struct ndis_work_item *ndis_work_item;
+};
+
 struct ndis_irq {
 	/* void *intr_obj is used for irq */
 	union {
@@ -702,6 +697,9 @@ struct ndis_miniport_block {
 	void *wan_tx_complete;
 	void *wan_rx;
 	void *wan_rx_complete;
+	/* ZyDas driver seems to corrupt memory without barrier below
+	 * - free_netdev results in slab error */
+	char barrier[10];
 	/* ndiswrapper specific */
 	struct wrap_ndis_device *wnd;
 };
@@ -722,16 +720,17 @@ struct wrap_ndis_device {
 	BOOLEAN stats_enabled;
 	struct ndis_wireless_stats ndis_stats;
 
-	struct work_struct xmit_work;
-	struct ndis_packet *xmit_ring[XMIT_RING_SIZE];
-	struct ndis_packet **xmit_array;
-	unsigned int xmit_ring_start;
-	unsigned int xmit_ring_pending;
-	unsigned int max_send_packets;
-	NT_SPIN_LOCK xmit_lock;
+	struct work_struct tx_work;
+	struct ndis_packet *tx_ring[TX_RING_SIZE];
+	struct ndis_packet **tx_array;
+	unsigned int tx_ring_start;
+	unsigned int tx_ring_end;
+	unsigned char is_tx_ring_full;
+	struct semaphore tx_ring_mutex;
+	unsigned int max_tx_packets;
 
-	unsigned char send_ok;
-	NT_SPIN_LOCK send_packet_done_lock;
+	unsigned char tx_ok;
+	NT_SPIN_LOCK tx_stats_lock;
 
 	struct semaphore ndis_comm_mutex;
 	wait_queue_head_t ndis_comm_wq;
@@ -768,8 +767,8 @@ struct wrap_ndis_device {
 	int iw_auth_cipher_group;
 	int iw_auth_key_mgmt;
 	int iw_auth_80211_auth_alg;
-	struct ndis_packet_pool *wrapper_packet_pool;
-	struct ndis_buffer_pool *wrapper_buffer_pool;
+	struct ndis_packet_pool *tx_packet_pool;
+	struct ndis_buffer_pool *tx_buffer_pool;
 };
 
 struct ndis_pmkid_candidate {
