@@ -16,7 +16,7 @@
 #ifndef _NTOSKERNEL_H_
 #define _NTOSKERNEL_H_
 
-#define UTILS_VERSION "1.7"
+#define UTILS_VERSION "1.8"
 
 #include <linux/types.h>
 #include <linux/timer.h>
@@ -63,8 +63,8 @@
 #define CONFIG_USB 1
 #endif
 
-#define addr_offset(drvr) (__builtin_return_address(0) - \
-			     (drvr)->drv_obj->driver_start)
+#define addr_offset(drvr) (__builtin_return_address(0) -	\
+			   (drvr)->drv_obj->driver_start)
 
 /* Workqueue / task queue backwards compatibility stuff */
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
@@ -197,7 +197,7 @@ typedef task_queue workqueue;
 #define PSMG_ON PM_ON
 #else
 typedef u32 pm_message_t;
-#define PMSG_SUSPEND 3
+#define PMSG_SUSPEND 2
 #define PMSG_ON 0
 #endif
 #endif
@@ -226,34 +226,98 @@ typedef u32 pm_message_t;
 #endif
 
 #ifdef CONFIG_X86_64
-#define LIN2WIN1(func, arg1)			\
-	lin_to_win1(func, (unsigned long)arg1)
+#define LIN2WIN1(func, arg1)						\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	lin_to_win1(func, (unsigned long)arg1);				\
+})
 #define LIN2WIN2(func, arg1, arg2)					\
-	lin_to_win2(func, (unsigned long)arg1, (unsigned long)arg2)
+({									\
+	DBGTRACE6("calling %p", func);					\
+	lin_to_win2(func, (unsigned long)arg1, (unsigned long)arg2);	\
+})
 #define LIN2WIN3(func, arg1, arg2, arg3)				\
+({									\
+	DBGTRACE6("calling %p", func);					\
 	lin_to_win3(func, (unsigned long)arg1, (unsigned long)arg2,	\
-		    (unsigned long)arg3)
+		    (unsigned long)arg3);				\
+})
 #define LIN2WIN4(func, arg1, arg2, arg3, arg4)				\
+({									\
+	DBGTRACE6("calling %p", func);					\
 	lin_to_win4(func, (unsigned long)arg1, (unsigned long)arg2,	\
-		    (unsigned long)arg3, (unsigned long)arg4)
+		    (unsigned long)arg3, (unsigned long)arg4);		\
+})
 #define LIN2WIN5(func, arg1, arg2, arg3, arg4, arg5)			\
+({									\
+	DBGTRACE6("calling %p", func);					\
 	lin_to_win5(func, (unsigned long)arg1, (unsigned long)arg2,	\
 		    (unsigned long)arg3, (unsigned long)arg4,		\
-		    (unsigned long)arg5)
+		    (unsigned long)arg5);				\
+})
 #define LIN2WIN6(func, arg1, arg2, arg3, arg4, arg5, arg6)		\
+({									\
+	DBGTRACE6("calling %p", func);					\
 	lin_to_win6(func, (unsigned long)arg1, (unsigned long)arg2,	\
 		    (unsigned long)arg3, (unsigned long)arg4,		\
-		    (unsigned long)arg5, (unsigned long)arg6)
-#else
-#define LIN2WIN1(func, arg1) func(arg1)
-#define LIN2WIN2(func, arg1, arg2) func(arg1, arg2)
-#define LIN2WIN3(func, arg1, arg2, arg3) func(arg1, arg2, arg3)
-#define LIN2WIN4(func, arg1, arg2, arg3, arg4) func(arg1, arg2, arg3, arg4)
-#define LIN2WIN5(func, arg1, arg2, arg3, arg4, arg5)	\
-	func(arg1, arg2, arg3, arg4, arg5)
-#define LIN2WIN6(func, arg1, arg2, arg3, arg4, arg5, arg6)	\
-	func(arg1, arg2, arg3, arg4, arg5, arg6)
-#endif
+		    (unsigned long)arg5, (unsigned long)arg6);		\
+})
+
+/* NOTE: these macros assume function arguments are quads and
+ * arguments are not touched in any way before calling these macros */
+#define WIN2LIN2(arg1, arg2)						\
+do {									\
+	__asm__ __volatile__("mov %%rcx, %0\n\t"			\
+			     "mov %%rdx, %1\n\t"			\
+			     : "=m" (arg1), "=m" (arg2));		\
+} while (0)
+
+#define WIN2LIN3(arg1, arg2, arg3)					\
+do {									\
+	__asm__ __volatile__("mov %%rcx, %0\n\t"			\
+			     "mov %%rdx, %1\n\t"			\
+			     "mov %%r8, %2\n\t"				\
+			     : "=m" (arg1), "=m" (arg2), "=m" (arg3));	\
+} while (0)
+
+#else // CONFIG_X86_64
+
+#define LIN2WIN1(func, arg1)						\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	func(arg1);							\
+})
+#define LIN2WIN2(func, arg1, arg2)					\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	func(arg1, arg2);						\
+})
+#define LIN2WIN3(func, arg1, arg2, arg3)				\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	func(arg1, arg2, arg3);						\
+})
+#define LIN2WIN4(func, arg1, arg2, arg3, arg4)				\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	func(arg1, arg2, arg3, arg4);					\
+})
+#define LIN2WIN5(func, arg1, arg2, arg3, arg4, arg5)			\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	func(arg1, arg2, arg3, arg4, arg5);				\
+})
+#define LIN2WIN6(func, arg1, arg2, arg3, arg4, arg5, arg6)		\
+({									\
+	DBGTRACE6("calling %p", func);					\
+	func(arg1, arg2, arg3, arg4, arg5, arg6);			\
+})
+
+#define WIN2LIN2(arg1, arg2) do { } while (0)
+
+#define WIN2LIN3(arg1, arg2, arg3) do { } while (0)
+
+#endif // CONFIG_X86_64
 
 #ifndef __wait_event_interruptible_timeout
 #define __wait_event_interruptible_timeout(wq, condition, ret)		\
@@ -340,6 +404,26 @@ do {									\
 #define WRAP_MODULE_PARM_STRING(name, perm) MODULE_PARM(name, "s")
 #endif
 
+#ifndef LOCK_PREFIX
+#ifdef LOCK
+#define LOCK_PREFIX LOCK
+#else
+#ifdef CONFIG_SMP
+#define LOCK_PREFIX "lock ; "
+#else
+#define LOCK_PREFIX ""
+#endif
+#endif
+#endif
+
+#ifndef NETDEV_TX_OK
+#define NETDEV_TX_OK 0
+#endif
+
+#ifndef NETDEV_TX_BUSY
+#define NETDEV_TX_BUSY 1
+#endif
+
 /* this ugly hack is to handle RH kernels; I don't know any better,
  * but this has to be fixed soon */
 #ifndef rt_task
@@ -364,7 +448,7 @@ do {									\
  * clock, otherwise from year 1601 */
 #define SYSTEM_TIME_TO_HZ(sys_time)					\
 	((((sys_time) <= 0) ? (((u64)HZ * (-(sys_time))) / TICKSPERSEC) : \
-	  (((u64)HZ * ((sys_time) - ticks_1601())) / TICKSPERSEC)))
+	  (((s64)HZ * ((sys_time) - ticks_1601())) / TICKSPERSEC)))
 
 #define MSEC_TO_HZ(ms) ((ms) * HZ / 1000)
 #define USEC_TO_HZ(ms) ((us) * HZ / 1000000)
@@ -408,7 +492,7 @@ struct wrap_alloc {
 
 struct pe_image {
 	char name[MAX_DRIVER_NAME_LEN];
-	void *entry;
+	UINT (*entry)(struct driver_object *, struct unicode_string *) STDCALL;
 	void *image;
 	int size;
 	int type;
@@ -432,16 +516,14 @@ struct wrap_timer {
 #endif
 };
 
-struct wrap_work_item {
+struct ntos_work_item {
 	struct nt_list list;
 	void *arg1;
 	void *arg2;
-	void *func;
+	void (*func)(void *arg1, void *arg2) STDCALL;
 	BOOLEAN win_func;
 };
 
-#define MAX_ALLOCATED_URBS 15
-
 struct wrap_device_setting {
 	struct nt_list list;
 	char name[MAX_SETTING_NAME_LEN];
@@ -468,9 +550,7 @@ struct wrap_driver {
 	int num_bin_files;
 	struct wrap_bin_file *bin_files;
 	struct nt_list wrap_devices;
-	union {
-		struct wrap_ndis_driver *ndis_driver;
-	};
+	struct wrap_ndis_driver *ndis_driver;
 };
 
 enum hw_status {
@@ -479,7 +559,6 @@ enum hw_status {
 };
 
 struct wrap_device {
-	struct nt_list list;
 	/* first part is (de)initialized once by loader */
 	int dev_bus_type;
 	int vendor;
@@ -492,8 +571,8 @@ struct wrap_device {
 	char conf_file_name[MAX_DRIVER_NAME_LEN];
 	struct nt_list settings;
 
-	/* rest should be (de)initialized during every
-	 * (de)initialization */
+	/* rest should be (de)initialized when a device is
+	 * (un)plugged */
 	struct device_object *pdo;
 	union {
 		struct {
@@ -507,24 +586,43 @@ struct wrap_device {
 			struct usb_interface *intf;
 			int num_alloc_urbs;
 			struct nt_list wrap_urb_list;
+			u8 reset;
 		} usb;
 	};
-	union {
-		struct wrap_ndis_device *wnd;
-	};
+	struct wrap_ndis_device *wnd;
 	struct nt_list timer_list;
 	NT_SPIN_LOCK timer_lock;
 	struct cm_resource_list *resource_list;
 };
 
+/* Some drivers use worker entries to complete functions called from
+ * within worker threads. So we should have separate workqueues to
+ * make sure worker entries run properly */
+
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
-/* until issues with threads hogging cpu are resolved, we don't want
- * to use shared workqueue, lest the threads take keyboard etc down */
 #define USE_OWN_WORKQUEUE 1
-extern struct workqueue_struct *wrapper_wq;
-#define schedule_work(work_struct) queue_work(wrapper_wq, (work_struct))
+extern struct workqueue_struct *wrap_wq;
+#define schedule_ndis_work(work_struct) queue_work(ndis_wq, (work_struct))
+#define schedule_wrap_work(work_struct) queue_work(wrap_wq, (work_struct))
+#else
+#define schedule_ndis_work(work_struct) schedule_work(work_struct)
+#define schedule_wrap_work(work_struct) schedule_work(work_struct)
 #endif
 
+/* Normally workqueue for ntos is not required, as worker entries in
+ * it are not supposed to wait; however, it helps to have separate
+ * workqueue so keyboard etc. work when kernel crashes */
+
+//#define USE_OWN_NTOS_WORKQUEUE 1
+
+#ifdef USE_OWN_NTOS_WORKQUEUE
+extern struct workqueue_struct *ntos_wq;
+#define schedule_ntos_work(work_struct) queue_work(ntos_wq, (work_struct))
+#else
+#define schedule_ntos_work(work_struct) schedule_work(work_struct)
+#endif
+
+
 int ntoskernel_init(void);
 void ntoskernel_exit(void);
 int ntoskernel_init_device(struct wrap_device *wd);
@@ -559,6 +657,9 @@ STDCALL ULONG MmSizeOfMdl(void *base, ULONG length);
 STDCALL void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
 			   enum memory_caching_type cache);
 STDCALL void MmUnmapIoSpace(void *addr, SIZE_T size);
+STDCALL void MmProbeAndLockPages(struct mdl *mdl, KPROCESSOR_MODE access_mode,
+	 enum lock_operation operation);
+STDCALL void MmUnlockPages(struct mdl *mdl);
 STDCALL void KeInitializeEvent(struct nt_event *nt_event,
 			       enum event_type type, BOOLEAN state);
 STDCALL LONG KeSetEvent(struct nt_event *nt_event, KPRIORITY incr,
@@ -614,7 +715,7 @@ STDCALL void KeInitializeEvent(struct nt_event *nt_event, enum event_type type,
 			       BOOLEAN state);
 void free_custom_extensions(struct driver_extension *drv_obj_ext);
 
-STDCALL struct irp *IoAllocateIrp(char stack_size, BOOLEAN charge_quota);
+STDCALL struct irp *IoAllocateIrp(char stack_count, BOOLEAN charge_quota);
 STDCALL void IoFreeIrp(struct irp *irp);
 STDCALL BOOLEAN IoCancelIrp(struct irp *irp);
 _FASTCALL NTSTATUS IofCallDriver
@@ -629,11 +730,10 @@ STDCALL struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
 	 struct io_status_block *status);
 STDCALL NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp);
 
-struct nt_thread *wrap_create_thread(struct task_struct *task);
-void wrap_remove_thread(struct nt_thread *thread);
+struct nt_thread *get_current_nt_thread(void);
 u64 ticks_1601(void);
 
-int schedule_wrap_work_item(WRAP_WORK_FUNC func, void *arg1, void *arg2,
+int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2,
 			    BOOLEAN win_func);
 
 STDCALL KIRQL KeGetCurrentIrql(void);
@@ -704,7 +804,8 @@ unsigned long lin_to_win6(void *func, unsigned long, unsigned long,
 			  unsigned long, unsigned long, unsigned long,
 			  unsigned long);
 
-STDCALL struct nt_thread *KeGetCurrentThread(void);
+
+STDCALL struct task_struct *KeGetCurrentThread(void);
 STDCALL NTSTATUS
 ObReferenceObjectByHandle(void *handle, ACCESS_MASK desired_access,
 			  void *obj_type, KPROCESSOR_MODE access_mode,
@@ -712,7 +813,6 @@ ObReferenceObjectByHandle(void *handle, ACCESS_MASK desired_access,
 
 _FASTCALL LONG ObfReferenceObject(FASTCALL_DECL_1(void *object));
 _FASTCALL void ObfDereferenceObject(FASTCALL_DECL_1(void *object));
-STDCALL NTSTATUS ZwClose(void *object);
 #define ObReferenceObject(object)			\
 	ObfReferenceObject(FASTCALL_ARGS_1(object))
 #define ObDereferenceObject(object)			\
@@ -738,21 +838,16 @@ void adjust_user_shared_data_addr(char *driver, unsigned long length);
 
 static inline KIRQL current_irql(void)
 {
-	if (in_atomic() || irqs_disabled())
+	if (in_irq() || irqs_disabled())
+		return DEVICE_LEVEL;
+	if (in_atomic())
 		return DISPATCH_LEVEL;
-	else
-		return PASSIVE_LEVEL;
+	return PASSIVE_LEVEL;
 }
 
 static inline KIRQL raise_irql(KIRQL newirql)
 {
 	KIRQL irql = current_irql();
-	/* for now we only deal with PASSIVE_LEVEL and
-	 * DISPATCH_LEVEL */
-#ifdef DEBUG_IRQL
-	if (newirql != DISPATCH_LEVEL)
-		ERROR("invalid irql: %d", newirql);
-#endif
 	if (irql < DISPATCH_LEVEL && newirql == DISPATCH_LEVEL) {
 		local_bh_disable();
 		preempt_disable();
@@ -762,6 +857,12 @@ static inline KIRQL raise_irql(KIRQL newirql)
 
 static inline void lower_irql(KIRQL oldirql)
 {
+#ifdef DEBUG_IRQL
+	KIRQL irql = current_irql();
+	if (irql < oldirql)
+		ERROR("invalid irql: %d < %d", irql, oldirql);
+#endif
+//	if (irql == DISPATCH_LEVEL && oldirql < DISPATCH_LEVEL) {
 	if (oldirql < DISPATCH_LEVEL) {
 		preempt_enable_no_resched();
 		local_bh_enable();
@@ -772,78 +873,77 @@ static inline void lower_irql(KIRQL oldirql)
  * store Linux spinlocks; so we implement Windows spinlocks using
  * ULONG_PTR space with our own functions/macros */
 
-/* the reason for value of unlocked spinlock to be 0, instead of 1
- * (which is what linux spinlocks use), is that some drivers don't
- * first call to initialize spinlock; in those case, the value of the
- * lock seems to be 0 (presumably in Windows value of unlocked
- * spinlock is 0).
- */
-
-/* define CONFIG_DEBUG_SPINLOCK if a Windows driver is suspected of
- * obtaining a lock while holding the same lock */
-
-//#ifndef CONFIG_DEBUG_SPINLOCK
-//#define CONFIG_DEBUG_SPINLOCK
-//#endif
-
-#undef CONFIG_DEBUG_SPINLOCK
-
-#ifdef CONFIG_DEBUG_SPINLOCK
-#define NT_SPIN_LOCK_LOCKED ((ULONG_PTR)current)
-#else
-#define NT_SPIN_LOCK_LOCKED 1
-#endif
+/* Windows seems to use 0 for unlocked state of spinlock - if Linux
+ * convention of 1 for unlocked state is used, at least prism54 driver
+ * crashes */
 
 #define NT_SPIN_LOCK_UNLOCKED 0
+#define NT_SPIN_LOCK_LOCKED 1
 
-#define nt_spin_lock_init(lock) *(lock) = NT_SPIN_LOCK_UNLOCKED
+static inline void  nt_spin_lock_init(volatile NT_SPIN_LOCK *lock)
+{
+	*lock = NT_SPIN_LOCK_UNLOCKED;
+}
 
 #ifdef CONFIG_SMP
+/* Linux kernel implements raw spinlocks in X86_64 differently from
+ * X86; as of now I understand following should work for both - if
+ * not, implement two versions */
+#if 0
+/* other spinlock implementations */
+	__asm__ __volatile__(
+		"\n"
+		"1:\t"
+		"  lock; xaddl %1, %0\n\t"
+		"  cmpl %2, %1\n\t"
+		"  je 3f\n"
+		"2:\t"
+		"  rep; nop\n\t"
+		"  cmpl %2, %0\n\t"
+		"  jne 2b\n\t"
+		"  mov $1, %1\n\t"
+		"  jmp 1b\n"
+		"3:\n\t"
+		: "=m" (*lock)
+		: "r" (1), "i" (NT_SPIN_LOCK_UNLOCKED)
+		: "memory");
+	while (xchg(lock, NT_SPIN_LOCK_LOCKED) != NT_SPIN_LOCK_UNLOCKED) {
+		do {
+			cpu_relax();
+		} while (*lock != NT_SPIN_LOCK_UNLOCKED);
+	}
+#endif
+static inline void nt_spin_lock(volatile NT_SPIN_LOCK *lock)
+{
+	__asm__ __volatile__(
+		"\n"
+		"1:\t"
+		"  xchgl %1, %0\n\t"
+		"  cmpl %2, %1\n\t"
+		"  je 3f\n"
+		"2:\t"
+		"  rep; nop\n\t"
+		"  cmpl %2, %0\n\t"
+		"  jne 2b\n\t"
+		"  jmp 1b\n"
+		"3:\n\t"
+		: "=m" (*lock)
+		: "r" (NT_SPIN_LOCK_LOCKED), "i" (NT_SPIN_LOCK_UNLOCKED)
+		: "memory");
+}
 
-#define raw_nt_spin_lock(lock)						\
-	while (cmpxchg((lock), NT_SPIN_LOCK_UNLOCKED, NT_SPIN_LOCK_LOCKED) != \
-	       NT_SPIN_LOCK_UNLOCKED)
+static inline void nt_spin_unlock(volatile NT_SPIN_LOCK *lock)
+{
+	*lock = NT_SPIN_LOCK_UNLOCKED;
+}
 
-#ifdef CONFIG_DEBUG_SPINLOCK
-#define raw_nt_spin_unlock(lock)					\
-	__asm__ __volatile__("movw $0,%0"				\
-			     :"=m" (*(lock)) : : "memory")
-#else // DEBUG_SPINLOCK
-#define raw_nt_spin_unlock(lock)					\
-	__asm__ __volatile__("movb $0,%0"				\
-			     :"=m" (*(lock)) : : "memory")
-#endif // DEBUG_SPINLOCK
+#else // CONFIG_SMP
 
-#else // SMP
+#define nt_spin_lock(lock) do { } while (0)
 
-#define raw_nt_spin_lock(lock) *(lock) = NT_SPIN_LOCK_LOCKED
-#define raw_nt_spin_unlock(lock) *(lock) = NT_SPIN_LOCK_UNLOCKED
+#define nt_spin_unlock(lock)  do { } while (0)
 
-#endif // SMP
-
-#ifdef CONFIG_DEBUG_SPINLOCK
-
-#define nt_spin_lock(lock)						\
-	do {								\
-		if (*(lock) == NT_SPIN_LOCK_LOCKED)			\
-			ERROR("eeek: process %p already owns lock %p",	\
-			      get_current(), lock);			\
-		else							\
-			raw_nt_spin_lock(lock);				\
-	} while (0)
-#define nt_spin_unlock(lock)						\
-	do {								\
-		if (*(lock) != NT_SPIN_LOCK_LOCKED)			\
-			ERROR("nt_spin_lock %p not locked!", (lock));	\
-		raw_nt_spin_unlock(lock);				\
-	} while (0)
-
-#else // DEBUG_SPINLOCK
-
-#define nt_spin_lock(lock) raw_nt_spin_lock(lock)
-#define nt_spin_unlock(lock) raw_nt_spin_unlock(lock)
-
-#endif // DEBUG_SPINLOCK
+#endif // CONFIG_SMP
 
 /* raise IRQL to given (higher) IRQL if necessary before locking */
 static inline KIRQL nt_spin_lock_irql(NT_SPIN_LOCK *lock, KIRQL newirql)
@@ -879,19 +979,111 @@ do {									\
 do {									\
 	nt_spin_unlock(lock);						\
 	restore_local_irq(flags);					\
-	preempt_enable_no_resched();					\
+	preempt_enable();						\
 } while (0)
 
-static inline ULONG SPAN_PAGES(ULONG_PTR ptr, SIZE_T length)
+static inline ULONG SPAN_PAGES(void *ptr, SIZE_T length)
 {
 	ULONG n;
 
-	n = (((ULONG_PTR)ptr & (PAGE_SIZE - 1)) +
+	n = ((((ULONG_PTR)ptr) & (PAGE_SIZE - 1)) +
 	     length + (PAGE_SIZE - 1)) >> PAGE_SHIFT;
 
 	return n;
 }
 
+#ifdef CONFIG_X86_64
+/* it is not clear how nt_slist_header is used to store pointer to
+ * slists and depth; here we assume 'align' field is used to store
+ * depth and 'region' field is used to store slist pointers */
+
+/* TODO: can these be implemented without using spinlock? */
+
+static inline struct nt_slist *PushEntrySList(nt_slist_header *head,
+					      struct nt_slist *entry,
+					      NT_SPIN_LOCK *lock)
+{
+	KIRQL irql = nt_spin_lock_irql(lock, DISPATCH_LEVEL);
+	entry->next = (struct nt_slist *)head->region;
+	head->region = (ULONGLONG)entry;
+	head->align = (head->align & 0x0000ffff) |
+		((head->align & 0xffff) + 1);
+	nt_spin_unlock_irql(lock, irql);
+	return entry->next;
+}
+
+static inline struct nt_slist *PopEntrySList(nt_slist_header *head,
+					     NT_SPIN_LOCK *lock)
+{
+	struct nt_slist *entry;
+	KIRQL irql = nt_spin_lock_irql(lock, DISPATCH_LEVEL);
+	entry = (struct nt_slist *)head->region;
+	if (entry) {
+		head->region = (ULONGLONG)entry->next;
+		head->align = (head->align & 0x0000ffff) |
+			((head->align & 0xffff) - 1);
+	}
+	nt_spin_unlock_irql(lock, irql);
+	return entry;
+}
+
+#else
+
+#ifndef ll_low
+#define ll_low(x) *(((u32 *)&(x)) + 0)
+#define ll_high(x) *(((u32 *)&(x)) + 1)
+#endif
+
+static inline u64 cmpxchg8b(volatile void *ptr, u64 old, u64 new)
+{
+	u64 prev;
+
+	__asm__ __volatile__(
+		"\n"
+		LOCK_PREFIX "cmpxchg8b (%0)\n"
+		: "+r" (ptr),
+		  "=a" (ll_low(prev)), "=d" (ll_high(prev))
+		: "a" (ll_low(old)), "d" (ll_high(old)),
+		  "b" (ll_low(new)), "c" (ll_high(new))
+		: "cc", "memory");
+	return prev;
+}
+
+/* slist routines below update slist atomically - no need for
+ * spinlocks */
+
+static inline struct nt_slist *PushEntrySList(nt_slist_header *head,
+					      struct nt_slist *entry,
+					      NT_SPIN_LOCK *lock)
+{
+	nt_slist_header old, new;
+	do {
+		old.align = head->align;
+		entry->next = old.next;
+		new.next = entry;
+		new.depth = old.depth + 1;
+	} while (cmpxchg8b(&head->align, old.align, new.align) != old.align);
+	return old.next;
+}
+
+static inline struct nt_slist *PopEntrySList(nt_slist_header *head,
+					     NT_SPIN_LOCK *lock)
+{
+	struct nt_slist *entry;
+	nt_slist_header old, new;
+	do {
+		old.align = head->align;
+		entry = old.next;
+		if (!entry)
+			break;
+		new.next = entry->next;
+		new.depth = old.depth - 1;
+	} while (cmpxchg8b(&head->align, old.align, new.align) != old.align);
+	return entry;
+}
+
+#endif
+
 /* DEBUG macros */
 
 #define DBGTRACE(fmt, ...) do { } while (0)
@@ -997,36 +1189,46 @@ do {								       \
 #define IOEXIT(stmt) stmt
 #endif
 
+#if defined(WORK_DEBUG)
+#define WORKTRACE DBGTRACE1
+#define WORKENTER TRACEENTER1
+#define WORKEXIT TRACEEXIT1
+#else
+#define WORKTRACE(fmt, ...)
+#define WORKENTER(fmt, ...)
+#define WORKEXIT(stmt) stmt
+#endif
+
 #if defined DEBUG
-#define assert(expr) do {						\
-		if (!(expr))						\
-			ERROR("assertion failed: %s", (#expr));		\
-	} while (0)
+#define assert(expr)							\
+do {									\
+	if (!(expr))							\
+		ERROR("assertion failed: %s", (#expr));			\
+} while (0)
 #else
 #define assert(expr) do { } while (0)
 #endif
 
 #if defined(IO_DEBUG)
 #define DUMP_IRP(__irp)							\
-	do {								\
-		struct io_stack_location *_irp_sl;			\
-		_irp_sl = IoGetCurrentIrpStackLocation(__irp);		\
-		IOTRACE("irp: %p, stack size: %d, cl: %d, sl: %p, "	\
-			"dev_obj: %p, mj_fn: %d, minor_fn: %d, "	\
-			"nt_urb: %p, event: %p",			\
-			__irp, __irp->stack_count, (__irp)->current_location, \
-			_irp_sl, _irp_sl->dev_obj, _irp_sl->major_fn,	\
-			_irp_sl->minor_fn, URB_FROM_IRP(__irp),		\
-			(__irp)->user_event);				\
-	} while (0)
+do {									\
+	struct io_stack_location *_irp_sl;				\
+	_irp_sl = IoGetCurrentIrpStackLocation(__irp);			\
+	IOTRACE("irp: %p, stack size: %d, cl: %d, sl: %p, dev_obj: %p, " \
+		"mj_fn: %d, minor_fn: %d, nt_urb: %p, event: %p",	\
+		__irp, __irp->stack_count, (__irp)->current_location,	\
+		_irp_sl, _irp_sl->dev_obj, _irp_sl->major_fn,		\
+		_irp_sl->minor_fn, URB_FROM_IRP(__irp),			\
+		(__irp)->user_event);					\
+} while (0)
 #else
 #define DUMP_IRP(__irp) do { } while (0)
 #endif
 
-#define sleep(nsec)					\
-	do {						\
-		set_current_state(TASK_INTERRUPTIBLE);	\
-		schedule_timeout(nsec * HZ);		\
-	} while (0)
+#define sleep_hz(n)					\
+do {							\
+	set_current_state(TASK_INTERRUPTIBLE);		\
+	schedule_timeout(n);				\
+} while (0)
 
 #endif // _NTOSKERNEL_H_
