@@ -21,7 +21,6 @@
 
 extern NT_SPIN_LOCK ntoskernel_lock;
 extern NT_SPIN_LOCK loader_lock;
-extern NT_SPIN_LOCK irp_cancel_lock;
 extern struct nt_list object_list;
 
 extern NT_SPIN_LOCK irp_cancel_lock;
@@ -29,15 +28,13 @@ extern NT_SPIN_LOCK irp_cancel_lock;
 STDCALL void WRAP_EXPORT(IoAcquireCancelSpinLock)
 	(KIRQL *irql)
 {
-	*irql = raise_irql(DISPATCH_LEVEL);
-	nt_spin_lock(&irp_cancel_lock);
+	*irql = nt_spin_lock_irql(&irp_cancel_lock, DISPATCH_LEVEL);
 }
 
 STDCALL void WRAP_EXPORT(IoReleaseCancelSpinLock)
 	(KIRQL irql)
 {
-	lower_irql(irql);
-	nt_spin_unlock(&irp_cancel_lock);
+	nt_spin_unlock_irql(&irp_cancel_lock, irql);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(IoGetDeviceProperty)
@@ -47,76 +44,37 @@ STDCALL NTSTATUS WRAP_EXPORT(IoGetDeviceProperty)
 {
 	struct ansi_string ansi;
 	struct unicode_string unicode;
-	char buf[32];
-	int devnum = 1;
+	struct wrap_device *wd;
+	ULONG need;
 
 	IOENTER("dev_obj = %p, dev_property = %d, buffer_len = %u, "
 		"buffer = %p, result_len = %p", pdo, dev_property,
 		buffer_len, buffer, result_len);
 
+	wd = pdo->reserved;
 	switch (dev_property) {
 	case DevicePropertyDeviceDescription:
-		if (buffer_len > 0 && buffer) {
-			*result_len = sizeof(int);
-			memset(buffer, 0xFF, *result_len);
-			IOEXIT(return STATUS_SUCCESS);
-		} else {
-			*result_len = sizeof(int);
-			IOEXIT(return STATUS_SUCCESS);
-		}
-		break;
-
 	case DevicePropertyFriendlyName:
-		if (buffer_len > 0 && buffer) {
-			ansi.length = snprintf(buf, sizeof(buf), "%d", devnum);
-			if (ansi.length <= 0) {
-				*result_len = 0;
-				IOEXIT(return STATUS_BUFFER_TOO_SMALL);
-			}
-			ansi.buf = buf;
-			ansi.max_length = ansi.length + 1;
-			unicode.buf = buffer;
-			unicode.max_length = buffer_len;
-			IOTRACE("unicode.length = %d, ansi.max_length = %d",
-				unicode.length, ansi.max_length);
-			if (RtlAnsiStringToUnicodeString(&unicode, &ansi,
-							 FALSE)) {
-				*result_len = 0;
-				IOEXIT(return STATUS_BUFFER_TOO_SMALL);
-			} else {
-				*result_len = unicode.max_length;
-				IOEXIT(return STATUS_SUCCESS);
-			}
-		} else {
-			ansi.max_length =
-				snprintf(buf, sizeof(buf), "%d", devnum);
-			*result_len = 2 * (ansi.max_length + 1);
+	case DevicePropertyDriverKeyName:
+		if (wrap_is_pci_bus(wd->dev_bus_type))
+			RtlInitAnsiString(&ansi, "PCI");
+		else // if (wrap_is_usb_bus(wd->dev_bus_type))
+			RtlInitAnsiString(&ansi, "USB");
+		need = sizeof(wchar_t) * (ansi.max_length + 1);
+		if (buffer_len < need) {
+			*result_len = need;
 			IOEXIT(return STATUS_BUFFER_TOO_SMALL);
 		}
-		break;
-
-	case DevicePropertyDriverKeyName:
-//		ansi.buf = wd->driver->name;
-		ansi.buf = buf;
-		ansi.length = strlen(ansi.buf);
-		ansi.max_length = ansi.length + 1;
-		if (buffer_len > 0 && buffer) {
-			unicode.buf = buffer;
-			unicode.length = buffer_len;
-			if (RtlAnsiStringToUnicodeString(&unicode, &ansi,
-							 FALSE)) {
-				*result_len = 0;
-				IOEXIT(return STATUS_BUFFER_TOO_SMALL);
-			} else {
-				*result_len = unicode.max_length;
-				IOEXIT(return STATUS_SUCCESS);
-			}
-		} else {
-				*result_len = 2 * (strlen(buf) + 1);
-				IOEXIT(return STATUS_SUCCESS);
+		unicode.max_length = buffer_len;
+		unicode.buf = buffer;
+		if (RtlAnsiStringToUnicodeString(&unicode, &ansi,
+						 FALSE) != STATUS_SUCCESS) {
+			*result_len = unicode.length;
+			IOEXIT(return STATUS_BUFFER_TOO_SMALL);
 		}
-		break;
+		IOEXIT(return STATUS_SUCCESS);
 	default:
+		WARNING("%d not implemented", dev_property);
 		IOEXIT(return STATUS_INVALID_PARAMETER_2);
 	}
 }
@@ -144,15 +102,17 @@ STDCALL BOOLEAN WRAP_EXPORT(IoIs32bitProcess)
 }
 
 STDCALL void WRAP_EXPORT(IoInitializeIrp)
-	(struct irp *irp, USHORT size, CCHAR stack_size)
+	(struct irp *irp, USHORT size, CCHAR stack_count)
 {
-	IOENTER("irp: %p, stack_size: %d", irp, stack_size);
+	IOENTER("irp: %p, stack_count: %d", irp, stack_count);
 
 	memset(irp, 0, size);
 	irp->size = size;
-	irp->stack_count = stack_size;
-	irp->current_location = stack_size + 1;
-	IoGetCurrentIrpStackLocation(irp) = IRP_SL(irp, (stack_size + 1));
+	irp->stack_count = stack_count;
+	/* IoAllocateIrp allocates space for one more than requested
+	 * stack_count */
+	irp->current_location = stack_count + 1;
+	IoGetCurrentIrpStackLocation(irp) = IRP_SL(irp, stack_count + 1);
 	IOEXIT(return);
 }
 
@@ -172,18 +132,20 @@ STDCALL void WRAP_EXPORT(IoReuseIrp)
 }
 
 STDCALL struct irp *WRAP_EXPORT(IoAllocateIrp)
-	(char stack_size, BOOLEAN charge_quota)
+	(char stack_count, BOOLEAN charge_quota)
 {
 	struct irp *irp;
 	int irp_size;
 
-	IOENTER("stack_size: %d, charge_quota: %d", stack_size, charge_quota);
-
-	irp_size = IoSizeOfIrp(stack_size + 1);
+	IOENTER("count: %d, quota: %d", stack_count, charge_quota);
+	/* driver need not allocate stack location for itself, but we
+	 * need to allocate space for it so that driver can set major
+	 * function etc. even if stack_count is 0 */
+	irp_size = IoSizeOfIrp(stack_count + 1);
 	irp = kmalloc(irp_size, GFP_ATOMIC);
 	if (irp) {
 		IOTRACE("allocated irp %p", irp);
-		IoInitializeIrp(irp, irp_size, stack_size);
+		IoInitializeIrp(irp, irp_size, stack_count);
 	}
 	IOEXIT(return irp);
 }
@@ -191,7 +153,7 @@ STDCALL struct irp *WRAP_EXPORT(IoAllocateIrp)
 STDCALL BOOLEAN WRAP_EXPORT(IoCancelIrp)
 	(struct irp *irp)
 {
-	void (*cancel_routine)(struct device_object *, struct irp *) STDCALL;
+	typeof(irp->cancel_routine) cancel_routine;
 
 	/* NB: this function may be called at DISPATCH_LEVEL */
 	IOENTER("irp = %p", irp);
@@ -201,6 +163,7 @@ STDCALL BOOLEAN WRAP_EXPORT(IoCancelIrp)
 	DUMP_IRP(irp);
 	IoAcquireCancelSpinLock(&irp->cancel_irql);
 	cancel_routine = irp->cancel_routine;
+	IOTRACE("%p", cancel_routine);
 	irp->cancel_routine = NULL;
 	irp->cancel = TRUE;
 	if (cancel_routine) {
@@ -208,9 +171,10 @@ STDCALL BOOLEAN WRAP_EXPORT(IoCancelIrp)
 
 		irp_sl = IoGetCurrentIrpStackLocation(irp);
 		/* cancel_routine will release the spin lock */
-		cancel_routine(irp_sl->dev_obj, irp);
+		LIN2WIN2(cancel_routine, irp_sl->dev_obj, irp);
 		IOEXIT(return TRUE);
 	} else {
+		IOTRACE("irp %p not canceled", irp);
 		IoReleaseCancelSpinLock(irp->cancel_irql);
 		IOEXIT(return FALSE);
 	}
@@ -221,19 +185,16 @@ STDCALL void IoQueueThreadIrp(struct irp *irp)
 	struct nt_thread *thread;
 	KIRQL irql;
 
-	thread = KeGetCurrentThread();
-	if (!thread) {
-		WARNING("couldn't find thread for irp: %p, task: %p, pid: %d",
-			irp, current, current->pid);
+	thread = get_current_nt_thread();
+	if (thread) {
+		IOTRACE("thread: %p, task: %p", thread, thread->task);
+		irql = nt_spin_lock_irql(&thread->lock, DISPATCH_LEVEL);
+		irp->flags |= IRP_SYNCHRONOUS_API;
+		InsertTailList(&thread->irps, &irp->threads);
+		IoIrpThread(irp) = thread;
+		nt_spin_unlock_irql(&thread->lock, irql);
+	} else
 		IoIrpThread(irp) = NULL;
-		return;
-	}
-	IOTRACE("thread: %p, task: %p", thread, thread->task);
-	irql = nt_spin_lock_irql(&thread->lock, DISPATCH_LEVEL);
-	irp->flags |= IRP_SYNCHRONOUS_API;
-	InsertTailList(&thread->irps, &irp->threads);
-	IoIrpThread(irp) = thread;
-	nt_spin_unlock_irql(&thread->lock, irql);
 }
 
 STDCALL void IoDequeueThreadIrp(struct irp *irp)
@@ -262,8 +223,7 @@ STDCALL void WRAP_EXPORT(IoFreeIrp)
 
 STDCALL struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
 	(ULONG major_fn, struct device_object *dev_obj, void *buffer,
-	 ULONG length, LARGE_INTEGER *offset,
-	 struct io_status_block *status)
+	 ULONG length, LARGE_INTEGER *offset, struct io_status_block *status)
 {
 	struct irp *irp;
 	struct io_stack_location *irp_sl;
@@ -271,9 +231,11 @@ STDCALL struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
 	IOENTER("%p", dev_obj);
 	if (!dev_obj)
 		IOEXIT(return NULL);
-	irp = IoAllocateIrp(dev_obj->stack_size, FALSE);
-	if (irp == NULL)
+	irp = IoAllocateIrp(dev_obj->stack_count, FALSE);
+	if (irp == NULL) {
+		WARNING("couldn't allocate irp");
 		IOEXIT(return NULL);
+	}
 
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	irp_sl->major_fn = major_fn;
@@ -286,24 +248,20 @@ STDCALL struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
 	irp_sl->completion_routine = NULL;
 
 	if (dev_obj->flags & DO_DIRECT_IO) {
-		IOTRACE("irp %p with DO_DIRECT_IO", irp);
 		irp->mdl = IoAllocateMdl(buffer, length, FALSE, FALSE, irp);
 		if (irp->mdl == NULL) {
 			IoFreeIrp(irp);
 			return NULL;
 		}
+		MmProbeAndLockPages(irp->mdl, KernelMode,
+				    major_fn == IRP_MJ_WRITE ?
+				    IoReadAccess : IoWriteAccess);
+		IOTRACE("mdl: %p", irp->mdl);
 	} else if (dev_obj->flags & DO_BUFFERED_IO) {
-		IOTRACE("irp %p with DO_BUFFERED_IO", irp);
-		irp->associated_irp.system_buffer =
-			ExAllocatePoolWithTag(NonPagedPool, length, 0);
-		if (irp->associated_irp.system_buffer == NULL) {
-			IoFreeIrp(irp);
-			return NULL;
-		}
-		irp->flags |= IRP_ASSOCIATED_IRP;
-		memcpy(irp->associated_irp.system_buffer, buffer, length);
-		irp->user_status = status;
-		irp->user_buf = buffer;
+		irp->associated_irp.system_buffer = buffer;
+		irp->flags = IRP_BUFFERED_IO;
+		irp->mdl = NULL;
+		IOTRACE("buffer: %p", buffer);
 	}
 	if (major_fn == IRP_MJ_READ) {
 		irp_sl->params.read.length = length;
@@ -344,26 +302,28 @@ STDCALL struct irp *WRAP_EXPORT(IoBuildDeviceIoControlRequest)
 	IOENTER("%p", dev_obj);
 	if (!dev_obj)
 		IOEXIT(return NULL);
-	irp = IoAllocateIrp(dev_obj->stack_size, FALSE);
-	if (irp) {
-		irp->user_status = io_status;
-		irp->user_event = event;
-		irp->user_buf = output_buf;
-		irp->associated_irp.system_buffer = input_buf;
-
-		irp_sl = IoGetNextIrpStackLocation(irp);
-		irp_sl->params.ioctl.code = ioctl;
-		irp_sl->params.ioctl.input_buf_len = input_buf_len;
-		irp_sl->params.ioctl.output_buf_len = output_buf_len;
-		irp_sl->dev_obj = dev_obj;
-		irp_sl->major_fn = (internal_ioctl) ?
-			IRP_MJ_INTERNAL_DEVICE_CONTROL : IRP_MJ_DEVICE_CONTROL;
-		irp_sl->minor_fn = 0;
-		irp_sl->flags = 0;
-		irp_sl->file_obj = NULL;
-		irp_sl->completion_routine = NULL;
-		IoQueueThreadIrp(irp);
+	irp = IoAllocateIrp(dev_obj->stack_count, FALSE);
+	if (irp == NULL) {
+		WARNING("couldn't allocate irp");
+		return NULL;
 	}
+	irp->user_status = io_status;
+	irp->user_event = event;
+	irp->user_buf = output_buf;
+	irp->associated_irp.system_buffer = input_buf;
+
+	irp_sl = IoGetNextIrpStackLocation(irp);
+	irp_sl->params.dev_ioctl.code = ioctl;
+	irp_sl->params.dev_ioctl.input_buf_len = input_buf_len;
+	irp_sl->params.dev_ioctl.output_buf_len = output_buf_len;
+	irp_sl->dev_obj = dev_obj;
+	irp_sl->major_fn = (internal_ioctl) ?
+		IRP_MJ_INTERNAL_DEVICE_CONTROL : IRP_MJ_DEVICE_CONTROL;
+	irp_sl->minor_fn = 0;
+	irp_sl->flags = 0;
+	irp_sl->file_obj = NULL;
+	irp_sl->completion_routine = NULL;
+	IoQueueThreadIrp(irp);
 
 	IOTRACE("irp: %p", irp);
 	IOEXIT(return irp);
@@ -373,7 +333,7 @@ _FASTCALL NTSTATUS WRAP_EXPORT(IofCallDriver)
 	(FASTCALL_DECL_2(struct device_object *dev_obj, struct irp *irp))
 {
 	struct io_stack_location *irp_sl;
-	NTSTATUS res;
+	NTSTATUS status;
 	driver_dispatch_t *major_func;
 	struct driver_object *drv_obj;
 
@@ -385,20 +345,21 @@ _FASTCALL NTSTATUS WRAP_EXPORT(IofCallDriver)
 	IOTRACE("drv_obj: %p", drv_obj);
 	major_func = drv_obj->major_func[irp_sl->major_fn];
 	IOTRACE("major_func: %p, dev_obj: %p", major_func, dev_obj);
+	/* TODO: Linux functions must be called natively */
 	if (major_func)
-		res = (*major_func)(dev_obj, irp);
+		status = LIN2WIN2(major_func, dev_obj, irp);
 	else {
 		ERROR("major_function %d is not implemented",
 		      irp_sl->major_fn);
-		res = STATUS_NOT_SUPPORTED;
+		status = STATUS_NOT_SUPPORTED;
 	}
-	IOEXIT(return res);
+	IOEXIT(return status);
 }
 
 _FASTCALL void WRAP_EXPORT(IofCompleteRequest)
 	(FASTCALL_DECL_2(struct irp *irp, CHAR prio_boost))
 {
-	NTSTATUS res;
+	NTSTATUS status;
 	struct io_stack_location *irp_sl;
 	struct mdl *mdl;
 
@@ -413,19 +374,21 @@ _FASTCALL void WRAP_EXPORT(IofCompleteRequest)
 		return;
 	}
 #endif
-	irp_sl = IoGetCurrentIrpStackLocation(irp);
-	/* completion routines expect current stack location to match
-	 * with what was set through IoSetCompletionRoutine, which
-	 * sets completion routine at next (higher) location, so we
-	 * skip current location and continue past stack size */
-	IoSkipCurrentIrpStackLocation(irp);
-
-	while (irp->current_location <= (irp->stack_count + 1)) {
+	for (irp_sl = IoGetCurrentIrpStackLocation(irp);
+	     irp->current_location <= irp->stack_count; irp_sl++) {
 		struct device_object *dev_obj;
 
 		if (irp_sl->control & SL_PENDING_RETURNED)
 			irp->pending_returned = TRUE;
 
+		/* current_location and dev_obj must be same as when
+		 * driver called IoSetCompletionRoutine, which sets
+		 * completion routine at next (lower) location, which
+		 * is what we are going to call below; so we set
+		 * current_location and dev_obj for the previous
+		 * (higher) location */
+		IoSkipCurrentIrpStackLocation(irp);
+
 		if (irp->current_location <= irp->stack_count)
 			dev_obj = IoGetCurrentIrpStackLocation(irp)->dev_obj;
 		else
@@ -433,29 +396,29 @@ _FASTCALL void WRAP_EXPORT(IofCompleteRequest)
 
 		if (irp_sl->completion_routine &&
 		    ((irp->io_status.status == STATUS_SUCCESS &&
-		       irp_sl->control & CALL_ON_SUCCESS) ||
+		       irp_sl->control & SL_INVOKE_ON_SUCCESS) ||
 		      (irp->io_status.status != STATUS_SUCCESS &&
-		       irp_sl->control & CALL_ON_ERROR) ||
-		      (irp->cancel && (irp_sl->control & CALL_ON_CANCEL)))) {
-			IOTRACE("calling completion_routine at: %p",
-				irp_sl->completion_routine);
-			res = LIN2WIN3(irp_sl->completion_routine, dev_obj,
-				       irp, irp_sl->context);
-			if (res == STATUS_MORE_PROCESSING_REQUIRED)
+		       irp_sl->control & SL_INVOKE_ON_ERROR) ||
+		      (irp->cancel == TRUE &&
+		       irp_sl->control & SL_INVOKE_ON_CANCEL))) {
+			IOTRACE("calling completion_routine at: %p, %p",
+				irp_sl->completion_routine, irp_sl->context);
+			status = LIN2WIN3(irp_sl->completion_routine,
+					  dev_obj, irp, irp_sl->context);
+			IOTRACE("status: %08X", status);
+			if (status == STATUS_MORE_PROCESSING_REQUIRED)
 				IOEXIT(return);
-			IOTRACE("completion routine returned");
 		} else {
+			/* propagate pending status to next irp_sl */
 			if (irp->current_location <= irp->stack_count &&
-			    irp->pending_returned)
+			    irp->pending_returned == TRUE)
 				IoMarkIrpPending(irp);
 		}
-		IoSkipCurrentIrpStackLocation(irp);
-		irp_sl++;
 	}
 
 	if (irp->user_status) {
 		irp->user_status->status = irp->io_status.status;
-		irp->user_status->status_info = irp->io_status.status_info;
+		irp->user_status->info = irp->io_status.info;
 	}
 
 	if (irp->user_event) {
@@ -464,12 +427,16 @@ _FASTCALL void WRAP_EXPORT(IofCompleteRequest)
 	}
 
 	IOTRACE("freeing irp %p", irp);
-	while ((mdl = irp->mdl)) {
-		irp->mdl = mdl->next;
-		IoFreeMdl(mdl);
-	}
-	if (irp->flags & IRP_DEALLOCATE_BUFFER)
+	if (irp->associated_irp.system_buffer &&
+	    (irp->flags & IRP_DEALLOCATE_BUFFER))
 		ExFreePool(irp->associated_irp.system_buffer);
+	else {
+		while ((mdl = irp->mdl)) {
+			irp->mdl = mdl->next;
+			MmUnlockPages(mdl);
+			IoFreeMdl(mdl);
+		}
+	}
 	IoFreeIrp(irp);
 	IOEXIT(return);
 }
@@ -479,22 +446,15 @@ static irqreturn_t io_irq_th(int irq, void *data, struct pt_regs *pt_regs)
 	struct kinterrupt *interrupt = (struct kinterrupt *)data;
 	NT_SPIN_LOCK *spinlock;
 	BOOLEAN ret;
-	KIRQL irql = PASSIVE_LEVEL;
 
 	if (interrupt->actual_lock)
 		spinlock = interrupt->actual_lock;
 	else
 		spinlock = &interrupt->lock;
-	if (interrupt->synch_irql >= DISPATCH_LEVEL)
-		irql = nt_spin_lock_irql(spinlock, DISPATCH_LEVEL);
-	else
-		nt_spin_lock(spinlock);
-	ret = interrupt->service_routine(interrupt,
-					 interrupt->service_context);
-	if (interrupt->synch_irql >= DISPATCH_LEVEL)
-		nt_spin_unlock_irql(spinlock, irql);
-	else
-		nt_spin_unlock(spinlock);
+	nt_spin_lock(spinlock);
+	ret = LIN2WIN2(interrupt->service_routine, interrupt,
+		       interrupt->service_context);
+	nt_spin_unlock(spinlock);
 
 	if (ret == TRUE)
 		return IRQ_HANDLED;
@@ -513,18 +473,14 @@ STDCALL NTSTATUS WRAP_EXPORT(IoConnectInterrupt)
 
 	interrupt->vector = vector;
 	interrupt->processor_enable_mask = processor_enable_mask;
-	nt_spin_lock_init(&interrupt->lock);
+//	nt_spin_lock_init(&interrupt->lock);
 	interrupt->actual_lock = lock;
 	interrupt->shareable = shareable;
 	interrupt->floating_save = floating_save;
 	interrupt->service_routine = service_routine;
 	interrupt->service_context = service_context;
 	InitializeListHead(&interrupt->list);
-	interrupt->irql = irql;
-	if (synch_irql > DISPATCH_LEVEL)
-		interrupt->synch_irql = DISPATCH_LEVEL;
-	else
-		interrupt->synch_irql = synch_irql;
+	interrupt->synch_irql = synch_irql;
 	interrupt->interrupt_mode = interrupt_mode;
 	if (request_irq(vector, io_irq_th, shareable ? SA_SHIRQ : 0,
 			"io_irq", interrupt)) {
@@ -589,15 +545,6 @@ STDCALL void WRAP_EXPORT(IoFreeWorkItem)
 	IOEXIT(return);
 }
 
-STDCALL void WRAP_EXPORT(ExQueueWorkItem)
-	(struct io_workitem *io_workitem, enum work_queue_type queue_type)
-{
-	IOENTER("%p", io_workitem);
-	schedule_wrap_work_item(io_workitem->worker_routine,
-				io_workitem->dev_obj, io_workitem->context,
-				TRUE);
-}
-
 STDCALL void WRAP_EXPORT(IoQueueWorkItem)
 	(struct io_workitem *io_workitem, void *func,
 	 enum work_queue_type queue_type, void *context)
@@ -605,10 +552,19 @@ STDCALL void WRAP_EXPORT(IoQueueWorkItem)
 	IOENTER("%p, %p", io_workitem, io_workitem->dev_obj);
 	io_workitem->worker_routine = func;
 	io_workitem->context = context;
-	schedule_wrap_work_item(func, io_workitem->dev_obj, context, TRUE);
+	schedule_ntos_work_item(func, io_workitem->dev_obj, context, TRUE);
 	IOEXIT(return);
 }
 
+STDCALL void WRAP_EXPORT(ExQueueWorkItem)
+	(struct io_workitem *io_workitem, enum work_queue_type queue_type)
+{
+	IOENTER("%p", io_workitem);
+	schedule_ntos_work_item(io_workitem->worker_routine,
+				io_workitem->dev_obj, io_workitem->context,
+				TRUE);
+}
+
 STDCALL NTSTATUS WRAP_EXPORT(IoAllocateDriverObjectExtension)
 	(struct driver_object *drv_obj, void *client_id, ULONG extlen,
 	 void **ext)
@@ -703,7 +659,7 @@ STDCALL NTSTATUS WRAP_EXPORT(IoCreateDevice)
 	dev->size = sizeof(*dev) + dev_ext_length;
 	dev->ref_count = 1;
 	dev->attached = NULL;
-	dev->stack_size = 1;
+	dev->stack_count = 1;
 
 	dev->drv_obj = drv_obj;
 	dev->next = drv_obj->dev_obj;
@@ -736,7 +692,7 @@ STDCALL NTSTATUS WRAP_EXPORT(IoCreateUnprotectedSymbolicLink)
 		IOTRACE("link: %s", ansi.buf);
 		RtlFreeAnsiString(&ansi);
 	}
-	UNIMPL();
+//	UNIMPL();
 	IOEXIT(return STATUS_SUCCESS);
 }
 
@@ -821,16 +777,16 @@ STDCALL struct device_object *WRAP_EXPORT(IoAttachDeviceToDeviceStack)
 
 	IOENTER("%p, %p", src, tgt);
 	dst = IoGetAttachedDevice(tgt);
-	IOTRACE("stack_size: %d -> %d", dst->stack_size, src->stack_size);
+	IOTRACE("stack_count: %d -> %d", dst->stack_count, src->stack_count);
 	IOTRACE("%p", dst);
 	irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 	if (dst)
 		dst->attached = src;
 	src->attached = NULL;
-	src->stack_size = dst->stack_size + 1;
+	src->stack_count = dst->stack_count + 1;
 //	src->ref_count++;
 	nt_spin_unlock_irql(&ntoskernel_lock, irql);
-	IOTRACE("stack_size: %d -> %d", dst->stack_size, src->stack_size);
+	IOTRACE("stack_count: %d -> %d", dst->stack_count, src->stack_count);
 	IOEXIT(return dst);
 }
 
@@ -853,22 +809,19 @@ STDCALL void WRAP_EXPORT(IoDetachDevice)
 	IOTRACE("attached:%p", topdev->attached);
 	for ( ; tail; tail = tail->attached) {
 		IOTRACE("tail:%p", tail);
-		tail->stack_size--;
+		tail->stack_count--;
 	}
 	nt_spin_unlock_irql(&ntoskernel_lock, irql);
 	IOEXIT(return);
 }
 
-/* gcc will return unions and structs either through registers or in
- * memory, so we can't return 'union power_state'; instead, we return
- * 'enum device_power_state' (which is one of the fields in the
- * union). With this gcc will honor STDCALL and return value
- * appropriately */
-STDCALL enum device_power_state WRAP_EXPORT(PoSetPowerState)
+/* NOTE: Make sure to compile with -freg-struct-return, so gcc will
+ * return union in register, like Windows */
+STDCALL union power_state WRAP_EXPORT(PoSetPowerState)
 	(struct device_object *dev_obj, enum power_state_type type,
 	 union power_state state)
 {
-	IOEXIT(return PowerDeviceD0);
+	IOEXIT(return state);
 }
 
 STDCALL NTSTATUS WRAP_EXPORT(PoCallDriver)
@@ -885,9 +838,9 @@ STDCALL NTSTATUS WRAP_EXPORT(PoRequestPowerIrp)
 	struct irp *irp;
 	struct io_stack_location *irp_sl;
 
-	DBGTRACE1("%p: stack size: %d", dev_obj, dev_obj->stack_size);
+	DBGTRACE1("%p: stack size: %d", dev_obj, dev_obj->stack_count);
 	DBGTRACE1("drv_obj: %p", dev_obj->drv_obj);
-	irp = IoAllocateIrp(dev_obj->stack_size, FALSE);
+	irp = IoAllocateIrp(dev_obj->stack_count, FALSE);
 	if (!irp)
 		return STATUS_INSUFFICIENT_RESOURCES;
 	irp_sl = IoGetNextIrpStackLocation(irp);
@@ -940,11 +893,11 @@ STDCALL NTSTATUS WRAP_EXPORT(IoRegisterDeviceInterface)
 {
 	struct ansi_string ansi;
 
-	/* check if pdo is valid */
-	ansi.buf = "ndis";
-	ansi.length = strlen(ansi.buf);
-	ansi.max_length = ansi.length + 1;
-	TRACEENTER1("pdo: %p, ref: %p, link: %p", pdo, reference, link);
+	/* TODO: check if pdo is valid */
+	RtlInitAnsiString(&ansi, "ndis");
+	TRACEENTER1("pdo: %p, ref: %p, link: %p, %x, %x, %x",
+		    pdo, reference, link, guid_class->data1,
+		    guid_class->data2, guid_class->data3);
 	return RtlAnsiStringToUnicodeString(link, &ansi, TRUE);
 }
 
@@ -964,49 +917,10 @@ STDCALL NTSTATUS WRAP_EXPORT(IoOpenDeviceRegistryKey)
 	return STATUS_SUCCESS;
 }
 
-STDCALL unsigned int WRAP_EXPORT(IoWMIRegistrationControl)
+STDCALL NTSTATUS WRAP_EXPORT(IoWMIRegistrationControl)
 	(struct device_object *dev_obj, ULONG action)
 {
-	struct irp *irp;
-	struct io_stack_location *irp_sl;
-	const int buf_len = 512;
-	void *buf;
-
 	TRACEENTER2("%p, %d", dev_obj, action);
-
-	switch (action) {
-	case WMIREG_ACTION_REGISTER:
-		irp = IoAllocateIrp(dev_obj->stack_size, FALSE);
-		if (!irp) {
-			ERROR("couldn't allocate irp");
-			return STATUS_INSUFFICIENT_RESOURCES;
-		}
-		irp_sl = IoGetNextIrpStackLocation(irp);
-		irp_sl->params.wmi.provider_id = (ULONG_PTR)dev_obj;
-		irp_sl->params.wmi.data_path = (void *)WMIREGISTER;
-		buf = kmalloc(buf_len, GFP_KERNEL);
-		if (!buf) {
-			IoFreeIrp(irp);
-			return STATUS_INSUFFICIENT_RESOURCES;
-		}
-		irp_sl->params.wmi.buf_len = buf_len;
-		irp_sl->params.wmi.buf = buf;
-		IoCallDriver(dev_obj, irp);
-		kfree(buf);
-		break;
-	case WMIREG_ACTION_DEREGISTER:
-		INFO("");
-		break;
-	case WMIREG_ACTION_REREGISTER:
-		INFO("");
-		break;
-	case WMIREG_ACTION_UPDATE_GUIDS:
-		ERROR("not implemented");
-		break;
-	default:
-		ERROR("not implemented");
-		break;
-	}
 	TRACEEXIT2(return STATUS_SUCCESS);
 }
 
