@@ -22,35 +22,64 @@ extern NT_SPIN_LOCK loader_lock;
 extern struct nt_list ndis_drivers;
 extern struct wrap_device *wrap_devices;
 
+/* IRP functions are called as Windows functions. For 64-bit, this
+ * means arguments are in rcx, rdx etc., but Linux functions expect
+ * them in rdi, rsi etc. So we need to put arguments back correctly
+ * before touching arguments. We also assume that all arguments are
+ * pointers (or register width). */
+
+/* called as Windows function, so call WIN2LIN3 before accessing
+ * arguments */
 STDCALL NTSTATUS IrpStopCompletion(struct device_object *dev_obj,
 				   struct irp *irp, void *context)
 {
+	WIN2LIN3(dev_obj, irp, context);
+
 	IOENTER("dev_obj: %p, irp: %p, context: %p", dev_obj, irp, context);
 	IOEXIT(return STATUS_MORE_PROCESSING_REQUIRED);
 }
 
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
+STDCALL NTSTATUS IopPassIrpDown(struct device_object *dev_obj,
+				struct irp *irp)
+{
+	WIN2LIN2(dev_obj, irp);
+
+	IoSkipCurrentIrpStackLocation(irp);
+	return IoCallDriver(dev_obj, irp);
+}
+
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
 STDCALL NTSTATUS IopInvalidDeviceRequest(struct device_object *dev_obj,
 					 struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	NTSTATUS status;
 
+	WIN2LIN2(dev_obj, irp);
+
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	WARNING("IRP %d:%d not implemented",
 		irp_sl->major_fn, irp_sl->minor_fn);
-	irp->io_status.status = STATUS_NOT_IMPLEMENTED;
-	irp->io_status.status_info = 0;
+	irp->io_status.status = STATUS_SUCCESS;
+	irp->io_status.info = 0;
 	status = irp->io_status.status;
 	IoCompleteRequest(irp, IO_NO_INCREMENT);
 	return status;
 }
 
-static STDCALL NTSTATUS pdoDispatchDeviceControl(struct device_object *pdo,
-						 struct  irp *irp)
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
+STDCALL NTSTATUS pdoDispatchDeviceControl(struct device_object *pdo,
+					  struct  irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	NTSTATUS status;
 
+	WIN2LIN2(pdo, irp);
+
 	DUMP_IRP(irp);
 
 	if (irp->current_location < 0 ||
@@ -58,7 +87,7 @@ static STDCALL NTSTATUS pdoDispatchDeviceControl(struct device_object *pdo,
 		ERROR("invalid irp: %p, %d, %d", irp, irp->current_location,
 		      irp->stack_count);
 		irp->io_status.status = STATUS_FAILURE;
-		irp->io_status.status_info = 0;
+		irp->io_status.info = 0;
 		IOEXIT(return STATUS_FAILURE);
 	}
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
@@ -66,25 +95,20 @@ static STDCALL NTSTATUS pdoDispatchDeviceControl(struct device_object *pdo,
 #ifdef CONFIG_USB
 	status = wrap_submit_irp(pdo, irp);
 	IOTRACE("status: %08X", status);
-	if (status == STATUS_PENDING) {
-		/* although as per DDK, we are not supposed to touch
-		 * irp when STAUS_PENDING is returned, this irp hasn't
-		 * been submitted to usb yet (and not completed), so
-		 * it is safe in this case */
-		status = wrap_submit_urb(irp);
-	}
 	if (status != STATUS_PENDING)
 		IoCompleteRequest(irp, IO_NO_INCREMENT);
 	IOEXIT(return status);
 #else
 	status = irp->io_status.status = STATUS_NOT_IMPLEMENTED;
 	IoCompleteRequest(irp, IO_NO_INCREMENT);
-	return status;
+	IOEXIT(return status);
 #endif
 }
 
-static STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
-				       struct irp *irp)
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
+STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
+				struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	struct wrap_device *wd;
@@ -93,6 +117,8 @@ static STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 	struct usbd_bus_interface_usbdi *usb_intf;
 #endif
 
+	WIN2LIN2(pdo, irp);
+
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wd = pdo->reserved;
 	DBGTRACE2("fn %d:%d, wd: %p", irp_sl->major_fn, irp_sl->minor_fn, wd);
@@ -166,8 +192,9 @@ static STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 	IOEXIT(return status);
 }
 
-static STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
-					 struct irp *irp)
+/* called as Windows function */
+STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
+				  struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	struct wrap_device *wd;
@@ -175,6 +202,8 @@ static STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
 	struct pci_dev *pdev;
 	NTSTATUS status;
 
+	WIN2LIN2(pdo, irp);
+
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wd = pdo->reserved;
 	DBGTRACE2("pdo: %p, fn: %d:%d, wd: %p",
@@ -199,6 +228,8 @@ static STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
 #else
 				pci_restore_state(pdev, wd->pci.pci_state);
 #endif
+			} else { // usb device
+				wrap_resume_urbs(wd);
 			}
 		} else {
 			DBGTRACE2("suspending device %p", wd);
@@ -209,6 +240,8 @@ static STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
 #else
 				pci_save_state(pdev, wd->pci.pci_state);
 #endif
+			} else { // usb device
+				wrap_suspend_urbs(wd);
 			}
 		}
 		status = STATUS_SUCCESS;
@@ -242,7 +275,8 @@ static struct device_object *alloc_pdo(struct driver_object *drv_obj)
 		drv_obj->major_func[i] = IopInvalidDeviceRequest;
 	drv_obj->major_func[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
 		pdoDispatchDeviceControl;
-	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] = pdoDispatchDeviceControl;
+	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] =
+		pdoDispatchDeviceControl;
 	drv_obj->major_func[IRP_MJ_POWER] = pdoDispatchPower;
 	drv_obj->major_func[IRP_MJ_PNP] = pdoDispatchPnp;
 	return pdo;
@@ -288,10 +322,10 @@ static int start_pdo(struct device_object *pdo)
 #ifdef CONFIG_X86_64
 	/* 64-bit broadcom driver doesn't work if DMA is allocated
 	 * from over 1GB */
-	if (strcmp(wd->driver->name, "netbc564") == 0) {
+	if (wd->vendor == 0x14e4) {
 		if (pci_set_dma_mask(pdev, 0x3fffffff) ||
 		    pci_set_consistent_dma_mask(pdev, 0x3fffffff))
-			WARNING("DMA mask couldn't be set; this driver "
+			WARNING("couldn't set DMA mask; this driver "
 				"may not work with more than 1GB RAM");
 	}
 #endif
@@ -303,9 +337,9 @@ static int start_pdo(struct device_object *pdo)
 			count++;
 	/* space for entry for IRQ is already in
 	 * cm_partial_resource_list */
-	resources_size = sizeof(struct cm_partial_resource_list) +
+	resources_size = sizeof(struct cm_resource_list) +
 		sizeof(struct cm_partial_resource_descriptor) * count;
-	DBGTRACE2("resources: %d, %d", i, resources_size);
+	DBGTRACE2("resources: %d, %d", count, resources_size);
 	wd->resource_list = kmalloc(resources_size, GFP_KERNEL);
 	if (!wd->resource_list) {
 		WARNING("couldn't allocate memory");
@@ -325,7 +359,7 @@ static int start_pdo(struct device_object *pdo)
 
 	for (i = count = 0; pci_resource_start(pdev, i); i++) {
 		entry = &partial_resource_list->partial_descriptors[count];
-		DBGTRACE2("flags: %lx", pci_resource_flags(pdev, i));
+		DBGTRACE2("%d", count);
 		if (pci_resource_flags(pdev, i) & IORESOURCE_MEM) {
 			entry->type = CmResourceTypeMemory;
 			entry->flags = CM_RESOURCE_MEMORY_READ_WRITE;
@@ -374,8 +408,10 @@ static int start_pdo(struct device_object *pdo)
 	entry->flags = CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE;
 	/* we assume all devices use shared IRQ */
 	entry->share = CmResourceShareShared;
-	/* 'level' should be DISPATCH_LEVEL, but some drivers, e.g.,
-	 * RTL8180L, use this also as vector, so set it to vector */
+	/* as per documentation, interrupt level should be DIRQL, but
+	 * examples from DDK as well some drivers, such as AR5211,
+	 * RT8180L use interrupt level as interrupt vector also in
+	 * NdisMRegisterInterrupt */
 	entry->u.interrupt.level = pdev->irq;
 	entry->u.interrupt.vector = pdev->irq;
 	entry->u.interrupt.affinity = -1;
@@ -399,20 +435,11 @@ NTSTATUS pnp_set_power_state(struct wrap_device *wd,
 	struct irp *irp;
 	struct io_stack_location *irp_sl;
 	NTSTATUS status;
-	struct nt_thread *thread;
 
-	if (KeGetCurrentThread() == NULL) {
-		thread = wrap_create_thread(current);
-		if (!thread) {
-			ERROR("couldn't allocate thread object");
-			TRACEEXIT1(return STATUS_FAILURE);
-		}
-	} else
-		thread = NULL;
 	pdo = wd->pdo;
 	fdo = IoGetAttachedDevice(pdo);
 	if (state > PowerDeviceD0) {
-		irp = IoAllocateIrp(fdo->stack_size, FALSE);
+		irp = IoAllocateIrp(fdo->stack_count, FALSE);
 		irp_sl = IoGetNextIrpStackLocation(irp);
 		DBGTRACE2("irp = %p, stack = %p", irp, irp_sl);
 		irp_sl->major_fn = IRP_MJ_POWER;
@@ -423,7 +450,7 @@ NTSTATUS pnp_set_power_state(struct wrap_device *wd,
 		if (status != STATUS_SUCCESS)
 			WARNING("query power returns %08X", status);
 	}
-	irp = IoAllocateIrp(fdo->stack_size, FALSE);
+	irp = IoAllocateIrp(fdo->stack_count, FALSE);
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	DBGTRACE2("irp = %p, stack = %p", irp, irp_sl);
 	irp_sl->major_fn = IRP_MJ_POWER;
@@ -433,8 +460,6 @@ NTSTATUS pnp_set_power_state(struct wrap_device *wd,
 	status = IoCallDriver(fdo, irp);
 	if (status != STATUS_SUCCESS)
 		WARNING("set power returns %08X", status);
-	if (thread)
-		wrap_remove_thread(thread);
 	TRACEEXIT1(return status);
 }
 
@@ -445,20 +470,11 @@ NTSTATUS pnp_start_device(struct wrap_device *wd)
 	struct irp *irp;
 	struct io_stack_location *irp_sl;
 	NTSTATUS status;
-	struct nt_thread *thread;
 
 	pdo = wd->pdo;
 	fdo = IoGetAttachedDevice(pdo);
 	DBGTRACE1("fdo: %p, irql: %d", fdo, current_irql());
-	if (KeGetCurrentThread() == NULL) {
-		thread = wrap_create_thread(current);
-		if (!thread) {
-			ERROR("couldn't allocate thread object");
-			TRACEEXIT1(return STATUS_FAILURE);
-		}
-	} else
-		thread = NULL;
-	irp = IoAllocateIrp(fdo->stack_size, FALSE);
+	irp = IoAllocateIrp(fdo->stack_count, FALSE);
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
 	/* TODO: for now we use same resources for both translated
@@ -476,8 +492,6 @@ NTSTATUS pnp_start_device(struct wrap_device *wd)
 	else
 		WARNING("Windows driver couldn't initialize the device (%08X)",
 			status);
-	if (thread)
-		wrap_remove_thread(thread);
 	TRACEEXIT1(return status);
 }
 
@@ -491,7 +505,7 @@ NTSTATUS pnp_stop_device(struct wrap_device *wd)
 	pdo = wd->pdo;
 	fdo = IoGetAttachedDevice(pdo);
 	DBGTRACE1("fdo: %p", fdo);
-	irp = IoAllocateIrp(fdo->stack_size, FALSE);
+	irp = IoAllocateIrp(fdo->stack_count, FALSE);
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
 	irp_sl->major_fn = IRP_MJ_PNP;
@@ -501,7 +515,7 @@ NTSTATUS pnp_stop_device(struct wrap_device *wd)
 	if (status != STATUS_SUCCESS)
 		WARNING("status: %08X", status);
 	/* for now we ignore query status */
-	irp = IoAllocateIrp(fdo->stack_size, FALSE);
+	irp = IoAllocateIrp(fdo->stack_count, FALSE);
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
 	irp_sl->major_fn = IRP_MJ_PNP;
@@ -519,20 +533,13 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 	struct driver_object *fdo_drv_obj;
 	struct irp *irp;
 	struct io_stack_location *irp_sl;
-	struct nt_thread *thread;
 	NTSTATUS status;
 
-	if (KeGetCurrentThread() == NULL) {
-		thread = wrap_create_thread(current);
-		if (!thread)
-			WARNING("couldn't allocate thread object");
-	} else
-		thread = NULL;
 	pdo = wd->pdo;
 	fdo = IoGetAttachedDevice(pdo);
 	fdo_drv_obj = fdo->drv_obj;
 	DBGTRACE1("fdo: %p", fdo);
-	irp = IoAllocateIrp(fdo->stack_size, FALSE);
+	irp = IoAllocateIrp(fdo->stack_count, FALSE);
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
 	irp_sl->major_fn = IRP_MJ_PNP;
@@ -542,7 +549,7 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 	if (status != STATUS_SUCCESS)
 		WARNING("status: %08X", status);
 	/* for now we ignore query status */
-	irp = IoAllocateIrp(fdo->stack_size, FALSE);
+	irp = IoAllocateIrp(fdo->stack_count, FALSE);
 	irp_sl = IoGetNextIrpStackLocation(irp);
 	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
 	irp_sl->major_fn = IRP_MJ_PNP;
@@ -551,7 +558,7 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 	status = IoCallDriver(fdo, irp);
 	if (status != STATUS_SUCCESS)
 		WARNING("status: %08X", status);
-	
+
 	/* TODO: should we use count in drv_ext or driver's Object
 	 * header reference count to keep count of devices associated
 	 * with a driver? */
@@ -576,8 +583,6 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 			ERROR("couldn't get wrap_driver");
 		ObDereferenceObject(fdo_drv_obj);
 	}
-	if (thread)
-		wrap_remove_thread(thread);
 	TRACEEXIT1(return status);
 }
 
@@ -606,7 +611,7 @@ static int wrap_pnp_start_device(struct wrap_device *wd)
 	driver = load_wrap_driver(wd);
 	if (!driver)
 		return -ENODEV;
-	
+
 	wd->driver = driver;
 	DBGTRACE1("dev type: %d, bus type: %d, %d",
 		  WRAP_DEVICE_TYPE(wd->dev_bus_type),
@@ -629,6 +634,7 @@ static int wrap_pnp_start_device(struct wrap_device *wd)
 			return -EINVAL;
 		}
 	}
+	DBGTRACE1("%p", driver->drv_obj->drv_ext->add_device);
 	if (driver->drv_obj->drv_ext->add_device(driver->drv_obj, pdo) !=
 	    STATUS_SUCCESS) {
 		IoDeleteDevice(pdo);
@@ -701,6 +707,27 @@ void *wrap_pnp_start_usb_device(struct usb_device *udev,
 	struct wrap_device *wd;
 	int ret;
 	wd = &wrap_devices[usb_id->driver_info];
+	DBGTRACE1("%04x, %04x, %x, %x, %x, %x, %x, %x, %x, %p",
+		  usb_id->idVendor, usb_id->idProduct,
+		  usb_id->bDeviceClass, usb_id->bDeviceSubClass,
+		  usb_id->bDeviceProtocol, usb_id->bInterfaceClass,
+		  usb_id->bDeviceProtocol, usb_id->bInterfaceClass,
+		  usb_id->bInterfaceSubClass, wd->usb.intf);
+
+	/* TI 4150 needs a full reset */
+	if (usb_id->idVendor == 0x057c && usb_id->idProduct == 0x5601 &&
+	    (xchg(&wd->usb.reset, 1) == 0)) {
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+		usb_set_intfdata(intf, NULL);
+		ret = usb_reset_device(interface_to_usbdev(intf));
+#else
+		ret = usb_reset_device(udev);
+#endif
+		if (ret) {
+			WARNING("reset failed: %d", ret);
+			goto out;
+		}
+	}
 	/* USB device may have multiple interfaces; initialize a
 	  device only once */
 	if (wd->usb.intf) {
@@ -717,6 +744,8 @@ void *wrap_pnp_start_usb_device(struct usb_device *udev,
 #endif
 		ret = wrap_pnp_start_device(wd);
 	}
+
+out:
 	DBGTRACE2("ret: %d", ret);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	if (ret)
@@ -774,7 +803,12 @@ int wrap_pnp_suspend_usb_device(struct usb_interface *intf, pm_message_t state)
 	if (!wd)
 		TRACEEXIT1(return -1);
 	pdo = wd->pdo;
-	return pnp_set_power_state(wd, PowerDeviceD3);
+	if (pnp_set_power_state(wd, PowerDeviceD3))
+		return -1;
+	else {
+//		intf->dev.power.power_state.event = state.event;
+		return 0;
+	}
 }
 
 int wrap_pnp_resume_usb_device(struct usb_interface *intf)
@@ -783,7 +817,12 @@ int wrap_pnp_resume_usb_device(struct usb_interface *intf)
 	wd = usb_get_intfdata(intf);
 	if (!wd)
 		TRACEEXIT1(return -1);
-	return pnp_set_power_state(wd, PowerDeviceD0);
+	if (pnp_set_power_state(wd, PowerDeviceD0))
+		return -1;
+	else {
+//		intf->dev.power.power_state.event = PM_EVENT_ON;
+		return 0;
+	}
 }
 #endif
 
