@@ -35,23 +35,10 @@ static void stats_timer_del(struct wrap_ndis_device *wnd);
 static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd);
 static NDIS_STATUS ndis_remove_device(struct wrap_ndis_device *wnd);
 
-static inline int ndis_wait_pending_completion(struct wrap_ndis_device *wnd)
+static inline int ndis_wait_comm_completion(struct wrap_ndis_device *wnd)
 {
-	/* wait for NdisMXXComplete to be called*/
-	/* As per spec, we should wait until miniport calls back
-	 * completion routine, but some drivers (e.g., ZyDas) don't
-	 * call back, so timeout is used; TODO: find out why drivers
-	 * don't call completion function */
-#if 1
-	/* setting PM state takes a long time, upto 2 seconds, for USB
-	 * devices */
-	if (wait_event_interruptible_timeout(wnd->ndis_comm_wq,
-					     (wnd->ndis_comm_done > 0),
-					     3 * HZ) <= 0)
-#else
-	if (wait_event_interruptible(wnd->ndis_comm_wq,
-				     (wnd->ndis_comm_done > 0)))
-#endif
+	if ((wait_event_interruptible(wnd->ndis_comm_wq,
+				      (wnd->ndis_comm_done > 0))))
 		return -1;
 	else
 		return 0;
@@ -86,7 +73,7 @@ NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd)
 	DBGTRACE2("res = %08X, reset_status = %08X", res, reset_address);
 	if (res == NDIS_STATUS_PENDING) {
 		/* wait for NdisMResetComplete */
-		if (ndis_wait_pending_completion(wnd))
+		if (ndis_wait_comm_completion(wnd))
 			res = NDIS_STATUS_FAILURE;
 		else {
 			res = wnd->ndis_comm_status;
@@ -136,11 +123,11 @@ NDIS_STATUS miniport_query_info_needed(struct wrap_ndis_device *wnd,
 	DBGTRACE2("res: %08X, oid: %08X", res, oid);
 	if (res == NDIS_STATUS_PENDING) {
 		/* wait for NdisMQueryInformationComplete */
-		if (ndis_wait_pending_completion(wnd))
+		if (ndis_wait_comm_completion(wnd))
 			res = NDIS_STATUS_FAILURE;
 		else
 			res = wnd->ndis_comm_status;
-		DBGTRACE2("res: %08X", res);
+		DBGTRACE2("res: %08X, oid: %08X", res, oid);
 	}
 	up(&wnd->ndis_comm_mutex);
 	if (res && needed)
@@ -186,11 +173,11 @@ NDIS_STATUS miniport_set_info(struct wrap_ndis_device *wnd, ndis_oid oid,
 	DBGTRACE2("res: %08X, oid: %08X", res, oid);
 	if (res == NDIS_STATUS_PENDING) {
 		/* wait for NdisMQueryInformationComplete */
-		if (ndis_wait_pending_completion(wnd))
+		if (ndis_wait_comm_completion(wnd))
 			res = NDIS_STATUS_FAILURE;
 		else
 			res = wnd->ndis_comm_status;
-		DBGTRACE2("res: %08X", res);
+		DBGTRACE2("res: %08X, oid: %08X", res, oid);
 	}
 	up(&wnd->ndis_comm_mutex);
 
@@ -212,6 +199,7 @@ NDIS_STATUS miniport_set_int(struct wrap_ndis_device *wnd, ndis_oid oid,
 	return miniport_set_info(wnd, oid, &data, sizeof(data));
 }
 
+/* MiniportInitialize */
 static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 {
 	NDIS_STATUS error_status, status;
@@ -224,11 +212,11 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 		ERROR("device %p already initialized!", wnd);
 		return NDIS_STATUS_FAILURE;
 	}
-	
+
 	if (!wnd->wd->driver->ndis_driver ||
 	    !wnd->wd->driver->ndis_driver->miniport.init) {
 		ERROR("assuming WDM (non-NDIS) driver");
-		TRACEEXIT1(return NDIS_STATUS_SUCCESS);
+		TRACEEXIT1(return NDIS_STATUS_NOT_RECOGNIZED);
 	}
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	status = LIN2WIN6(miniport->init, &error_status,
@@ -244,7 +232,7 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 	/* Wait a little to let card power up otherwise ifup might
 	 * fail after boot; USB devices seem to need long delays */
 	set_current_state(TASK_INTERRUPTIBLE);
-	schedule_timeout(HZ);
+	schedule_timeout(1 * HZ);
 
 #if 0
 	status = miniport_set_pm_state(wnd, NdisDeviceStateD0);
@@ -259,6 +247,7 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 	TRACEEXIT1(return NDIS_STATUS_SUCCESS);
 }
 
+/* MiniportHalt */
 static void miniport_halt(struct wrap_ndis_device *wnd)
 {
 	struct miniport_char *miniport;
@@ -273,14 +262,13 @@ static void miniport_halt(struct wrap_ndis_device *wnd)
 	}
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	DBGTRACE1("driver halt is at %p", miniport->miniport_halt);
-
-	DBGTRACE2("task: %p, pid: %d", current, current->pid);
 	LIN2WIN1(miniport->miniport_halt, wnd->nmb->adapter_ctx);
 	clear_bit(HW_INITIALIZED, &wnd->hw_status);
 
 	TRACEEXIT1(return);
 }
 
+/* MiniportPnPEventNotify */
 static NDIS_STATUS miniport_pnp_event(struct wrap_ndis_device *wnd,
 				      enum ndis_device_pnp_event event)
 {
@@ -347,7 +335,7 @@ static void ndis_set_multicast_list(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	set_bit(SET_MULTICAST_LIST, &wnd->wrap_ndis_work);
-	schedule_work(&wnd->wrap_ndis_worker);
+	schedule_wrap_work(&wnd->wrap_ndis_worker);
 }
 
 static int ndis_set_mac_addr(struct net_device *dev, void *p)
@@ -431,29 +419,33 @@ static void show_supported_oids(struct wrap_ndis_device *wnd)
 	TRACEEXIT1(return);
 }
 
-static struct ndis_packet *
-allocate_send_packet(struct wrap_ndis_device *wnd, ndis_buffer *buffer)
+static struct ndis_packet *alloc_tx_packet(struct wrap_ndis_device *wnd,
+					   struct sk_buff *skb)
 {
 	struct ndis_packet *packet;
+	ndis_buffer *buffer;
 	struct ndis_packet_oob_data *oob_data;
 	NDIS_STATUS status;
 
-	NdisAllocatePacket(&status, &packet, wnd->wrapper_packet_pool);
+	NdisAllocatePacket(&status, &packet, wnd->tx_packet_pool);
 	if (status != NDIS_STATUS_SUCCESS)
 		return NULL;
-
+	NdisAllocateBuffer(&status, &buffer, wnd->tx_buffer_pool,
+			   skb->data, skb->len);
+	if (status != NDIS_STATUS_SUCCESS) {
+		NdisFreePacket(packet);
+		return NULL;
+	}
 	packet->private.buffer_head = buffer;
 	packet->private.buffer_tail = buffer;
 
 	oob_data = NDIS_PACKET_OOB_DATA(packet);
+	oob_data->skb = skb;
 	if (wnd->use_sg_dma) {
 		oob_data->ndis_sg_element.address =
-			PCI_DMA_MAP_SINGLE(wnd->wd->pci.pdev,
-					   MmGetMdlVirtualAddress(buffer),
-					   MmGetMdlByteCount(buffer),
-					   PCI_DMA_TODEVICE);
-
-		oob_data->ndis_sg_element.length = MmGetMdlByteCount(buffer);
+			PCI_DMA_MAP_SINGLE(wnd->wd->pci.pdev, skb->data,
+					   skb->len, PCI_DMA_TODEVICE);
+		oob_data->ndis_sg_element.length = skb->len;
 		oob_data->ndis_sg_list.nent = 1;
 		oob_data->ndis_sg_list.elements = &oob_data->ndis_sg_element;
 		oob_data->extension.info[ScatterGatherListPacketInfo] =
@@ -462,189 +454,208 @@ allocate_send_packet(struct wrap_ndis_device *wnd, ndis_buffer *buffer)
 	return packet;
 }
 
-static void free_send_packet(struct wrap_ndis_device *wnd,
-			     struct ndis_packet *packet)
+void free_tx_packet(struct wrap_ndis_device *wnd, struct ndis_packet *packet,
+		    NDIS_STATUS status)
 {
 	ndis_buffer *buffer;
 	struct ndis_packet_oob_data *oob_data;
+	KIRQL irql;
 
-	TRACEENTER3("packet: %p", packet);
-	buffer = packet->private.buffer_head;
+	TRACEENTER3("%p, %08X", packet, status);
+	irql = nt_spin_lock_irql(&wnd->tx_stats_lock, DISPATCH_LEVEL);
+	if (status == NDIS_STATUS_SUCCESS) {
+		wnd->stats.tx_bytes += packet->private.len;
+		wnd->stats.tx_packets++;
+	} else {
+		WARNING("packet dropped: %08X", status);
+		wnd->stats.tx_dropped++;
+	}
+	nt_spin_unlock_irql(&wnd->tx_stats_lock, irql);
 	oob_data = NDIS_PACKET_OOB_DATA(packet);
 	if (wnd->use_sg_dma)
 		PCI_DMA_UNMAP_SINGLE(wnd->wd->pci.pdev,
 				     oob_data->ndis_sg_element.address,
 				     oob_data->ndis_sg_element.length,
 				     PCI_DMA_TODEVICE);
-
+	buffer = packet->private.buffer_head;
 	DBGTRACE3("freeing buffer %p", buffer);
 	NdisFreeBuffer(buffer);
-	dev_kfree_skb_any(oob_data->skb);
-
+	if (oob_data->skb)
+		dev_kfree_skb_any(oob_data->skb);
 	DBGTRACE3("freeing packet %p", packet);
 	NdisFreePacket(packet);
 	TRACEEXIT3(return);
 }
 
-/*
- * MiniportSend and MiniportSendPackets
- * this function is called with lock held in DISPATCH_LEVEL, so no need
- * to raise irql to DISPATCH_LEVEL during MiniportSend(Packets)
-*/
-static int send_packets(struct wrap_ndis_device *wnd, unsigned int start,
-			unsigned int pending)
+/* MiniportSend and MiniportSendPackets */
+/* this function is called holding tx_ring_mutex, so safe to read
+ * tx_ring_start (tx_ring_end is not updated in tx_worker or here, so
+ * safe to read tx_ring_end, too) without lock */
+static int miniport_tx_packets(struct wrap_ndis_device *wnd)
 {
 	NDIS_STATUS res;
 	struct miniport_char *miniport;
-	unsigned int sent, n;
+	int n, sent, start, end;
 	struct ndis_packet *packet;
+	KIRQL irql;
 
-	TRACEENTER3("start: %d, pending: %d", start, pending);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
-	if (pending > wnd->max_send_packets)
-		n = wnd->max_send_packets;
-	else
-		n = pending;
-
+	start = wnd->tx_ring_start;
+	end = wnd->tx_ring_end;
+	/* end == start when ring is full: (TX_RING_SIZE - 1) number
+	 * of packets are pending */
+	n = end - start;
+	if (n < 0)
+		n += TX_RING_SIZE;
+	else if (n == 0) {
+		assert(wnd->is_tx_ring_full == 1);
+		n = TX_RING_SIZE - 1;
+	}
+	if (unlikely(n > wnd->max_tx_packets))
+		n = wnd->max_tx_packets;
+	DBGTRACE3("%d, %d, %d", n, start, end);
 	if (miniport->send_packets) {
 		int i;
-		/* copy packets from xmit ring to linear xmit array */
+		/* copy packets from tx ring to linear tx array */
 		for (i = 0; i < n; i++) {
-			int j = (start + i) % XMIT_RING_SIZE;
-			wnd->xmit_array[i] = wnd->xmit_ring[j];
+			int j = (start + i) % TX_RING_SIZE;
+			wnd->tx_array[i] = wnd->tx_ring[j];
 		}
-		LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
-			 wnd->xmit_array, n);
-		DBGTRACE3("sent");
 		if (test_bit(ATTR_SERIALIZED, &wnd->attributes)) {
-			for (sent = 0; sent < n && wnd->send_ok; sent++) {
+			irql = raise_irql(DISPATCH_LEVEL);
+			LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
+				 wnd->tx_array, n);
+			lower_irql(irql);
+			for (sent = 0; sent < n && wnd->tx_ok; sent++) {
 				struct ndis_packet_oob_data *oob_data;
-				packet = wnd->xmit_array[sent];
+				packet = wnd->tx_array[sent];
 				oob_data = NDIS_PACKET_OOB_DATA(packet);
-				switch(oob_data->status) {
+				switch (oob_data->status) {
 				case NDIS_STATUS_SUCCESS:
-					sendpacket_done(wnd, packet);
+					free_tx_packet(wnd, packet,
+						       oob_data->status);
 					break;
 				case NDIS_STATUS_PENDING:
 					break;
 				case NDIS_STATUS_RESOURCES:
-					wnd->send_ok = 0;
+					DBGTRACE3("%d, %d", start, end);
+					wnd->tx_ok = 0;
+					/* resubmit this packet and
+					 * the rest when resources
+					 * become available */
+					netif_stop_queue(wnd->net_dev);
+					sent--;
 					break;
 				case NDIS_STATUS_FAILURE:
+					WARNING("packet %p dropped", packet);
+					free_tx_packet(wnd, packet,
+						       oob_data->status);
+					break;
 				default:
-					free_send_packet(wnd, packet);
+					ERROR("packet %p: unknown status %08X",
+					      packet, oob_data->status);
+					/* RT2500 driver seems to
+					 * corrupt packets with SMP
+					 * and by the time control
+					 * comes here, the buffer and
+					 * packet are already freed */
 					break;
 				}
 			}
 		} else {
+			LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
+				 wnd->tx_array, n);
 			sent = n;
 		}
+		DBGTRACE3("sent: %d(%d)", sent, n);
 	} else {
-		packet = wnd->xmit_ring[start];
-		res = LIN2WIN3(miniport->send, wnd->nmb->adapter_ctx,
-			       packet, 0);
+		packet = wnd->tx_ring[start];
+		if (test_bit(ATTR_SERIALIZED, &wnd->attributes)) {
+			irql = raise_irql(DISPATCH_LEVEL);
+			res = LIN2WIN3(miniport->send, wnd->nmb->adapter_ctx,
+				       packet, packet->private.flags);
+			lower_irql(irql);
+		} else
+			res = LIN2WIN3(miniport->send, wnd->nmb->adapter_ctx,
+				       packet, packet->private.flags);
 		sent = 1;
 		switch (res) {
 		case NDIS_STATUS_SUCCESS:
-			sendpacket_done(wnd, packet);
+			free_tx_packet(wnd, packet, res);
 			break;
 		case NDIS_STATUS_PENDING:
 			break;
 		case NDIS_STATUS_RESOURCES:
-			wnd->send_ok = 0;
+			wnd->tx_ok = 0;
 			sent = 0;
 			break;
 		case NDIS_STATUS_FAILURE:
-			free_send_packet(wnd, packet);
+			free_tx_packet(wnd, packet, res);
 			break;
 		}
 	}
 	TRACEEXIT3(return sent);
 }
 
-static void xmit_worker(void *param)
+static void tx_worker(void *param)
 {
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)param;
 	int n;
-	KIRQL irql;
 
-	TRACEENTER3("send_ok %d", wnd->send_ok);
+	TRACEENTER3("tx_ok %d", wnd->tx_ok);
 
-	/* some drivers e.g., new RT2500 driver, crash if any packets
-	 * are sent when the card is not associated */
-	while (wnd->send_ok) {
-		irql = nt_spin_lock_irql(&wnd->xmit_lock, DISPATCH_LEVEL);
-		if (wnd->xmit_ring_pending == 0) {
-			nt_spin_unlock_irql(&wnd->xmit_lock, irql);
+	/* we can use cmpxchg to update tx_ring_start, but we need to
+	 * send packets in given order, so serialize (even for
+	 * deserialized drivers) with mutex */
+
+	while (1) {
+		if (down_interruptible(&wnd->tx_ring_mutex))
+			break;
+		/* end == start if either ring is empty or full; in
+		 * the latter case is_tx_ring_full is set */
+		if ((wnd->tx_ring_end == wnd->tx_ring_start &&
+		     !wnd->is_tx_ring_full) || !wnd->tx_ok) {
+			up(&wnd->tx_ring_mutex);
 			break;
 		}
-		n = send_packets(wnd, wnd->xmit_ring_start,
-				 wnd->xmit_ring_pending);
-		wnd->xmit_ring_start =
-			(wnd->xmit_ring_start + n) % XMIT_RING_SIZE;
-		wnd->xmit_ring_pending -= n;
-		if (netif_queue_stopped(wnd->net_dev) && n > 0)
-			netif_wake_queue(wnd->net_dev);
-		nt_spin_unlock_irql(&wnd->xmit_lock, irql);
+		n = miniport_tx_packets(wnd);
+		if (n > 0) {
+			wnd->tx_ring_start =
+				(wnd->tx_ring_start + n) % TX_RING_SIZE;
+			wnd->is_tx_ring_full = 0;
+			if (netif_queue_stopped(wnd->net_dev))
+				netif_wake_queue(wnd->net_dev);
+		}
+		up(&wnd->tx_ring_mutex);
+		DBGTRACE3("%d, %d, %d", n, wnd->tx_ring_start,
+			  wnd->tx_ring_end);
 	}
-
 	TRACEEXIT3(return);
 }
 
-/*
- * Free and unmap packet created in xmit
- */
-void sendpacket_done(struct wrap_ndis_device *wnd,
-		     struct ndis_packet *packet)
-{
-
-	TRACEENTER3("%p", packet);
-	nt_spin_lock(&wnd->send_packet_done_lock);
-	wnd->stats.tx_bytes += packet->private.len;
-	wnd->stats.tx_packets++;
-	nt_spin_unlock(&wnd->send_packet_done_lock);
-	free_send_packet(wnd, packet);
-	TRACEEXIT3(return);
-}
-
-/*
- * This function is called in BH disabled context and ndis drivers
- * must have their send-functions called from sleepeable context so we
- * just queue the packets up here and schedule a workqueue to run
- * later.
- */
-static int start_xmit(struct sk_buff *skb, struct net_device *dev)
+static int tx_skbuff(struct sk_buff *skb, struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	ndis_buffer *buffer;
 	struct ndis_packet *packet;
-	struct ndis_packet_oob_data *oob_data;
-	unsigned int xmit_ring_next_slot;
-	NDIS_STATUS res;
 
-	NdisAllocateBuffer(&res, &buffer, wnd->wrapper_buffer_pool,
-			   skb->data, skb->len);
-	if (res != NDIS_STATUS_SUCCESS)
-		return 1;
-	packet = allocate_send_packet(wnd, buffer);
+	packet = alloc_tx_packet(wnd, skb);
 	if (!packet) {
-		NdisFreeBuffer(buffer);
-		return 1;
+		WARNING("couldn't allocate packet");
+		return NETDEV_TX_BUSY;
 	}
-	oob_data = NDIS_PACKET_OOB_DATA(packet);
-	oob_data->skb = skb;
-	nt_spin_lock(&wnd->xmit_lock);
-	xmit_ring_next_slot = (wnd->xmit_ring_start +
-			       wnd->xmit_ring_pending) % XMIT_RING_SIZE;
-	wnd->xmit_ring[xmit_ring_next_slot] = packet;
-	wnd->xmit_ring_pending++;
-	if (wnd->xmit_ring_pending == XMIT_RING_SIZE)
+	/* no need for lock here - already called holding
+	 * net_dev->xmit_lock and tx_ring_end is not updated
+	 * elsewhere */
+	wnd->tx_ring[wnd->tx_ring_end++] = packet;
+	if (wnd->tx_ring_end == TX_RING_SIZE)
+		wnd->tx_ring_end = 0;
+	if (wnd->tx_ring_end == wnd->tx_ring_start) {
+		wnd->is_tx_ring_full = 1;
 		netif_stop_queue(wnd->net_dev);
-	nt_spin_unlock(&wnd->xmit_lock);
-
-	schedule_work(&wnd->xmit_work);
-
-	return 0;
+	}
+	DBGTRACE3("%d, %d", wnd->tx_ring_start, wnd->tx_ring_end);
+	schedule_wrap_work(&wnd->tx_work);
+	return NETDEV_TX_OK;
 }
 
 static int set_packet_filter(struct wrap_ndis_device *wnd, ULONG packet_filter)
@@ -689,6 +700,7 @@ static int ndis_open(struct net_device *dev)
 	if (dev->flags & IFF_PROMISC)
 		packet_filter |= NDIS_PACKET_TYPE_PROMISCUOUS |
 			NDIS_PACKET_TYPE_ALL_LOCAL;
+	/* NDIS_PACKET_TYPE_PROMISCUOUS may not work with 802.11 */
 	if (set_packet_filter(wnd, packet_filter) &&
 	    packet_filter & NDIS_PACKET_TYPE_PROMISCUOUS) {
 		DBGTRACE1("couldn't add packet filter %x", packet_filter);
@@ -697,7 +709,6 @@ static int ndis_open(struct net_device *dev)
 			DBGTRACE1("couldn't add packet filter %x",
 				  packet_filter);
 	}
-	/* NDIS_PACKET_TYPE_PROMISCUOUS will not work with 802.11 */
 	netif_device_attach(dev);
 	netif_start_queue(dev);
 	return 0;
@@ -791,43 +802,15 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 	unsigned char *assoc_info;
 	union iwreq_data wrqu;
 	NDIS_STATUS res;
-	const int assoc_size = sizeof(*ndis_assoc_info) + IW_CUSTOM_MAX;
-
+	const int assoc_size = sizeof(*ndis_assoc_info) + IW_CUSTOM_MAX + 32;
 	TRACEENTER2("link status: %d", wnd->link_status);
 	if (wnd->link_status == 0) {
-#if 0
-		unsigned int i;
-		struct encr_info *encr_info = &wnd->encr_info;
-
-		if (wnd->encr_mode == Ndis802_11Encryption1Enabled ||
-		    wnd->infrastructure_mode == Ndis802_11IBSS) {
-			for (i = 0; i < MAX_ENCR_KEYS; i++) {
-				if (encr_info->keys[i].length == 0)
-					continue;
-				add_wep_key(wnd, encr_info->keys[i].key,
-					    encr_info->keys[i].length, i);
-			}
-
-			set_bit(SET_ESSID, &wnd->wrap_ndis_work);
-			schedule_work(&wnd->wrap_ndis_worker);
-			TRACEEXIT2(return);
-		}
-		/* TODO: not clear if NDIS says keys should
-		 * be cleared here */
-		for (i = 0; i < MAX_ENCR_KEYS; i++)
-			wnd->encr_info.keys[i].length = 0;
-#endif
-
 		memset(&wrqu, 0, sizeof(wrqu));
 		wrqu.ap_addr.sa_family = ARPHRD_ETHER;
 		wireless_send_event(wnd->net_dev, SIOCGIWAP, &wrqu, NULL);
 		TRACEEXIT2(return);
 	}
 
-	if (!(test_bit(Ndis802_11Encryption2Enabled, &wnd->capa.encr) ||
-	      test_bit(Ndis802_11Encryption3Enabled, &wnd->capa.encr)))
-		TRACEEXIT2(return);
-
 	assoc_info = kmalloc(assoc_size, GFP_KERNEL);
 	if (!assoc_info) {
 		ERROR("couldn't allocate memory");
@@ -836,14 +819,6 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 	memset(assoc_info, 0, assoc_size);
 
 	ndis_assoc_info = (struct ndis_assoc_info *)assoc_info;
-#if 0
-	ndis_assoc_info->length = sizeof(*ndis_assoc_info);
-	ndis_assoc_info->offset_req_ies = sizeof(*ndis_assoc_info);
-	ndis_assoc_info->req_ie_length = IW_CUSTOM_MAX / 2;
-	ndis_assoc_info->offset_resp_ies = sizeof(*ndis_assoc_info) +
-		ndis_assoc_info->req_ie_length;
-	ndis_assoc_info->resp_ie_length = IW_CUSTOM_MAX / 2;
-#endif
 	res = miniport_query_info(wnd, OID_802_11_ASSOCIATION_INFORMATION,
 				  assoc_info, assoc_size);
 	if (res) {
@@ -922,7 +897,7 @@ static void stats_proc(unsigned long data)
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)data;
 
 	set_bit(COLLECT_STATS, &wnd->wrap_ndis_work);
-	schedule_work(&wnd->wrap_ndis_worker);
+	schedule_wrap_work(&wnd->wrap_ndis_worker);
 	wnd->stats_timer.expires += 10 * HZ;
 	add_timer(&wnd->stats_timer);
 }
@@ -947,7 +922,7 @@ static void hangcheck_proc(unsigned long data)
 
 	TRACEENTER3("");
 	set_bit(HANGCHECK, &wnd->wrap_ndis_work);
-	schedule_work(&wnd->wrap_ndis_worker);
+	schedule_wrap_work(&wnd->wrap_ndis_worker);
 
 	wnd->hangcheck_timer.expires += wnd->hangcheck_interval;
 	add_timer(&wnd->hangcheck_timer);
@@ -1183,10 +1158,11 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 		if (test_and_clear_bit(HW_HALTED, &wnd->hw_status)) {
 			DBGTRACE2("starting device");
 			/* TODO: should we use pnp_start_device instead? */
-			if (miniport_init(wnd) != NDIS_STATUS_SUCCESS) {
+			status = miniport_init(wnd);
+			if (status != NDIS_STATUS_SUCCESS) {
 				WARNING("couldn't re-initialize device %s",
 					wnd->net_dev->name);
-				TRACEEXIT2(return NDIS_STATUS_FAILURE);
+				TRACEEXIT2(return status);
 			}
 		}
 		if (test_and_clear_bit(HW_SUSPENDED, &wnd->hw_status)) {
@@ -1218,19 +1194,25 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 	}
 }
 
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
 STDCALL NTSTATUS NdisDispatchDeviceControl(struct device_object *fdo,
 					   struct irp *irp)
 {
 	struct wrap_ndis_device *wnd;
+
+	WIN2LIN2(fdo, irp);
+
 	DBGTRACE3("fdo: %p", fdo);
 	/* for now, we don't have anything intresting here, so pass it
 	 * down to bus driver */
 	wnd = fdo->reserved;
-	return IopPassIrpDown(wnd->nmb->pdo, irp);
+	return LIN2WIN2(IopPassIrpDown, wnd->nmb->pdo, irp);
 }
 
-STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo,
-				   struct irp *irp)
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
+STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	struct wrap_ndis_device *wnd;
@@ -1238,6 +1220,8 @@ STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo,
 	NTSTATUS status;
 	NDIS_STATUS ndis_status;
 
+	WIN2LIN2(fdo, irp);
+
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wnd = fdo->reserved;
 	IOTRACE("fdo: %p, fn: %d:%d, wnd: %p", fdo, irp_sl->major_fn,
@@ -1254,7 +1238,7 @@ STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo,
 					       TRUE, FALSE, FALSE);
 			status = IoCallDriver(wnd->nmb->pdo, irp);
 			if (status != STATUS_SUCCESS)
-				return status;
+				break;
 			ndis_status = miniport_set_power_state(wnd, state);
 			if (ndis_status != NDIS_STATUS_SUCCESS)
 				WARNING("setting power to %d failed: %08X",
@@ -1262,7 +1246,6 @@ STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo,
 			hangcheck_add(wnd);
 			stats_timer_add(wnd);
 			set_scan(wnd);
-			set_bit(SET_ESSID, &wnd->wrap_ndis_work);
 			if (netif_running(wnd->net_dev)) {
 				netif_device_attach(wnd->net_dev);
 				netif_wake_queue(wnd->net_dev);
@@ -1288,7 +1271,7 @@ STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo,
 			if (ndis_status != NDIS_STATUS_SUCCESS)
 				WARNING("setting power to %d failed: %08X",
 					state, ndis_status);
-			return IopPassIrpDown(wnd->nmb->pdo, irp);
+			return LIN2WIN2(IopPassIrpDown, wnd->nmb->pdo, irp);
 		}
 		status = STATUS_SUCCESS;
 		break;
@@ -1308,22 +1291,25 @@ STDCALL NTSTATUS NdisDispatchPower(struct device_object *fdo,
 		status = STATUS_NOT_SUPPORTED;
 		break;
 	default:
-		return IopPassIrpDown(wnd->nmb->pdo, irp);
+		return LIN2WIN2(IopPassIrpDown, wnd->nmb->pdo, irp);
 	}
-	irp->io_status.status_info = 0;
+	irp->io_status.info = 0;
 	irp->io_status.status = status;
 	IoCompleteRequest(irp, IO_NO_INCREMENT);
 	return status;
 }
 
-STDCALL NTSTATUS NdisDispatchPnp(struct device_object *fdo,
-				 struct irp *irp)
+/* called as Windows function, so call WIN2LIN2 before accessing
+ * arguments */
+STDCALL NTSTATUS NdisDispatchPnp(struct device_object *fdo, struct irp *irp)
 {
 	struct io_stack_location *irp_sl;
 	struct wrap_ndis_device *wnd;
 	struct device_object *pdo;
 	NTSTATUS status;
 
+	WIN2LIN2(fdo, irp);
+
 	IOENTER("fdo: %p, irp: %p", fdo, irp);
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	wnd = fdo->reserved;
@@ -1344,19 +1330,19 @@ STDCALL NTSTATUS NdisDispatchPnp(struct device_object *fdo,
 			status = STATUS_FAILURE;
 		break;
 	case IRP_MN_QUERY_STOP_DEVICE:
-		return IopPassIrpDown(pdo, irp);
+		return LIN2WIN2(IopPassIrpDown, pdo, irp);
 	case IRP_MN_STOP_DEVICE:
 		miniport_halt(wnd);
-		return IopPassIrpDown(pdo, irp);
+		return LIN2WIN2(IopPassIrpDown, pdo, irp);
 	case IRP_MN_REMOVE_DEVICE:
 		DBGTRACE1("%s", wnd->net_dev->name);
 		if (ndis_remove_device(wnd) != NDIS_STATUS_SUCCESS) {
 			status = STATUS_FAILURE;
 			break;
 		}
-		return IopPassIrpDown(pdo, irp);
+		return LIN2WIN2(IopPassIrpDown, pdo, irp);
 	default:
-		return IopPassIrpDown(pdo, irp);
+		return LIN2WIN2(IopPassIrpDown, pdo, irp);
 	}
 	irp->io_status.status = status;
 	IOTRACE("res: %08X", status);
@@ -1370,13 +1356,16 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	struct net_device *net_dev;
 	NDIS_STATUS ndis_status;
 	mac_address mac;
-	char buf[256];
+	char buf[64];
 
-	if (miniport_init(wnd) != NDIS_STATUS_SUCCESS)
-		return NDIS_STATUS_FAILURE;
-	/* NB: xmit_array is used to recognize if device is being
+	ndis_status = miniport_init(wnd);
+	if (ndis_status == NDIS_STATUS_NOT_RECOGNIZED)
+		TRACEEXIT1(return NDIS_STATUS_SUCCESS);
+	if (ndis_status != NDIS_STATUS_SUCCESS)
+		TRACEEXIT1(return ndis_status);
+	/* NB: tx_array is used to recognize if device is being
 	 * started for the first time or being re-started */
-	if (wnd->xmit_array)
+	if (wnd->tx_array)
 		TRACEEXIT2(return NDIS_STATUS_SUCCESS);
 	wd = wnd->wd;
 	net_dev = wnd->net_dev;
@@ -1394,8 +1383,10 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	DBGTRACE1("mac:" MACSTR, MAC2STR(mac));
 	memcpy(&net_dev->dev_addr, mac, ETH_ALEN);
 
+	check_capa(wnd);
+
 	net_dev->open = ndis_open;
-	net_dev->hard_start_xmit = start_xmit;
+	net_dev->hard_start_xmit = tx_skbuff;
 	net_dev->stop = ndis_close;
 	net_dev->get_stats = ndis_get_stats;
 	net_dev->do_ioctl = ndis_ioctl;
@@ -1424,11 +1415,11 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	if (ndis_status == NDIS_STATUS_SUCCESS)
 		printk(KERN_INFO "%s: vendor: '%s'\n", net_dev->name, buf);
 
-	printk(KERN_INFO "%s: %s ethernet device " MACSTR " using driver %s,"
+	printk(KERN_INFO "%s: %s ethernet device " MACSTR " using %sdriver %s,"
 	       " %s\n", net_dev->name, DRIVER_NAME, MAC2STR(net_dev->dev_addr),
+	       test_bit(ATTR_SERIALIZED, &wnd->attributes) ? "serialized " : "",
 	       wnd->wd->driver->name, wnd->wd->conf_file_name);
 
-	check_capa(wnd);
 	DBGTRACE1("capbilities = %ld", wnd->capa.encr);
 	printk(KERN_INFO "%s: encryption modes supported: %s%s%s%s%s%s%s\n",
 	       net_dev->name,
@@ -1452,66 +1443,65 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	if (test_bit(ATTR_SERIALIZED, &wnd->attributes)) {
 		ndis_status =
 			miniport_query_int(wnd, OID_GEN_MAXIMUM_SEND_PACKETS,
-					   &wnd->max_send_packets);
+					   &wnd->max_tx_packets);
 		if (ndis_status == NDIS_STATUS_NOT_SUPPORTED)
-			wnd->max_send_packets = 1;
-		if (wnd->max_send_packets > XMIT_RING_SIZE)
-			wnd->max_send_packets = XMIT_RING_SIZE;
+			wnd->max_tx_packets = 1;
+		if (wnd->max_tx_packets > TX_RING_SIZE)
+			wnd->max_tx_packets = TX_RING_SIZE;
 	} else {
 		/* deserialized drivers don't have a limit, but we
-		 * keep max at XMIT_RING_SIZE to allocate xmit_array
+		 * keep max at TX_RING_SIZE to allocate tx_array
 		 * below */
-		wnd->max_send_packets = XMIT_RING_SIZE;
+		wnd->max_tx_packets = TX_RING_SIZE;
 	}
-
-	DBGTRACE1("maximum send packets: %d", wnd->max_send_packets);
-	wnd->xmit_array =
-		kmalloc(sizeof(struct ndis_packet *) * wnd->max_send_packets,
+	DBGTRACE1("maximum send packets: %d", wnd->max_tx_packets);
+	wnd->tx_array =
+		kmalloc(sizeof(struct ndis_packet *) * wnd->max_tx_packets,
 			GFP_KERNEL);
-	if (!wnd->xmit_array) {
+	if (!wnd->tx_array) {
 		ERROR("couldn't allocate memory for tx_packets");
 		goto err_start;
 
 	}
 	/* we need at least one extra packet for
 	 * EthRxIndicateHandler */
-	NdisAllocatePacketPoolEx(&ndis_status, &wnd->wrapper_packet_pool,
-				 wnd->max_send_packets + 1, 0,
+	NdisAllocatePacketPoolEx(&ndis_status, &wnd->tx_packet_pool,
+				 wnd->max_tx_packets + 1, 0,
 				 PROTOCOL_RESERVED_SIZE_IN_PACKET);
 	if (ndis_status != NDIS_STATUS_SUCCESS) {
 		ERROR("couldn't allocate packet pool");
 		goto packet_pool_err;
 	}
-	NdisAllocateBufferPool(&ndis_status, &wnd->wrapper_buffer_pool,
-			       wnd->max_send_packets + 4);
+	NdisAllocateBufferPool(&ndis_status, &wnd->tx_buffer_pool,
+			       wnd->max_tx_packets + 4);
 	if (ndis_status != NDIS_STATUS_SUCCESS) {
 		ERROR("couldn't allocate buffer pool");
 		goto buffer_pool_err;
 	}
-	DBGTRACE1("pool: %p", wnd->wrapper_buffer_pool);
+	DBGTRACE1("pool: %p", wnd->tx_buffer_pool);
 	miniport_set_int(wnd, OID_802_11_NETWORK_TYPE_IN_USE,
 			 Ndis802_11Automode);
-//	miniport_set_int(wnd, OID_802_11_POWER_MODE, NDIS_POWER_OFF);
+	miniport_set_int(wnd, OID_802_11_POWER_MODE, NDIS_POWER_OFF);
 	/* check_capa changes auth_mode and encr_mode, so set them again */
 	set_infra_mode(wnd, Ndis802_11Infrastructure);
+	set_scan(wnd);
+	set_priv_filter(wnd, Ndis802_11PrivFilterAcceptAll);
 	set_auth_mode(wnd, Ndis802_11AuthModeOpen);
 	set_encr_mode(wnd, Ndis802_11EncryptionDisabled);
-	set_scan(wnd);
-	set_privacy_filter(wnd, Ndis802_11PrivFilterAcceptAll);
 	set_essid(wnd, "", 0);
 
 	wrap_procfs_add_ndis_device(wnd);
 	TRACEEXIT1(return NDIS_STATUS_SUCCESS);
 
 buffer_pool_err:
-	wnd->wrapper_buffer_pool = NULL;
-	if (wnd->wrapper_packet_pool) {
-		NdisFreePacketPool(wnd->wrapper_packet_pool);
-		wnd->wrapper_packet_pool = NULL;
+	wnd->tx_buffer_pool = NULL;
+	if (wnd->tx_packet_pool) {
+		NdisFreePacketPool(wnd->tx_packet_pool);
+		wnd->tx_packet_pool = NULL;
 	}
 packet_pool_err:
-	kfree(wnd->xmit_array);
-	wnd->xmit_array = NULL;
+	kfree(wnd->tx_array);
+	wnd->tx_array = NULL;
 err_start:
 	ndis_remove_device(wnd);
 	TRACEEXIT1(return NDIS_STATUS_FAILURE);
@@ -1519,43 +1509,52 @@ err_start:
 
 static NDIS_STATUS ndis_remove_device(struct wrap_ndis_device *wnd)
 {
-	KIRQL irql;
-
+	int tx_pending;
 	set_bit(SHUTDOWN, &wnd->wrap_ndis_work);
-	miniport_pnp_event(wnd, NdisDevicePnPEventSurpriseRemoved);
+	wnd->tx_ok = 0;
 	ndis_close(wnd->net_dev);
 	netif_carrier_off(wnd->net_dev);
-	wrap_procfs_remove_ndis_device(wnd);
+	/* In 2.4 kernels, this function is called in atomic context,
+	 * so we can't/don't need to wait on mutex. */
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+	cancel_delayed_work(&wnd->wrap_ndis_worker);
+	down_interruptible(&wnd->tx_ring_mutex);
+#endif
+	tx_pending = wnd->tx_ring_end - wnd->tx_ring_start;
+	if (tx_pending < 0)
+		tx_pending += TX_RING_SIZE;
+	else if (tx_pending == 0 && wnd->is_tx_ring_full)
+		tx_pending = TX_RING_SIZE - 1;
+	wnd->is_tx_ring_full = 0;
 	/* throw away pending packets */
-	irql = nt_spin_lock_irql(&wnd->xmit_lock, DISPATCH_LEVEL);
-	while (wnd->xmit_ring_pending) {
+	while (tx_pending > 0) {
 		struct ndis_packet *packet;
 
-		packet = wnd->xmit_ring[wnd->xmit_ring_start];
-		free_send_packet(wnd, packet);
-		wnd->xmit_ring_start =
-			(wnd->xmit_ring_start + 1) % XMIT_RING_SIZE;
-		wnd->xmit_ring_pending--;
+		packet = wnd->tx_ring[wnd->tx_ring_start];
+		free_tx_packet(wnd, packet, NDIS_STATUS_CLOSING);
+		wnd->tx_ring_start = (wnd->tx_ring_start + 1) % TX_RING_SIZE;
+		tx_pending--;
 	}
-	nt_spin_unlock_irql(&wnd->xmit_lock, irql);
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+	up(&wnd->tx_ring_mutex);
+#endif
+	miniport_pnp_event(wnd, NdisDevicePnPEventSurpriseRemoved);
+	wrap_procfs_remove_ndis_device(wnd);
 	miniport_halt(wnd);
 	ndis_exit_device(wnd);
-	/* flush_scheduled_work here causes crash with 2.4 kernels */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-	flush_scheduled_work();
-#endif
-	if (wnd->wrapper_packet_pool) {
-		NdisFreePacketPool(wnd->wrapper_packet_pool);
-		wnd->wrapper_packet_pool = NULL;
+
+	if (wnd->tx_packet_pool) {
+		NdisFreePacketPool(wnd->tx_packet_pool);
+		wnd->tx_packet_pool = NULL;
 	}
-	if (wnd->wrapper_buffer_pool) {
-		NdisFreeBufferPool(wnd->wrapper_buffer_pool);
-		wnd->wrapper_buffer_pool = NULL;
+	if (wnd->tx_buffer_pool) {
+		NdisFreeBufferPool(wnd->tx_buffer_pool);
+		wnd->tx_buffer_pool = NULL;
 	}
 	IoDetachDevice(wnd->nmb->next_device);
 	IoDeleteDevice(wnd->nmb->fdo);
-	if (wnd->xmit_array)
-		kfree(wnd->xmit_array);
+	if (wnd->tx_array)
+		kfree(wnd->tx_array);
 	printk(KERN_INFO "%s: device %s removed\n", DRIVER_NAME,
 	       wnd->net_dev->name);
 	unregister_netdev(wnd->net_dev);
@@ -1572,7 +1571,7 @@ static STDCALL NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	struct wrap_ndis_device *wnd;
 	struct net_device *net_dev;
 	struct wrap_device *wd;
-	int i;
+	unsigned long i;
 
 	TRACEENTER2("%p, %p", drv_obj, pdo);
 	if (strlen(if_name) > (IFNAMSIZ-1)) {
@@ -1603,8 +1602,15 @@ static STDCALL NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 
 	nmb = ((void *)wnd) + sizeof(*wnd);
 	wnd->nmb = nmb;
+#if defined(DEBUG) && DEBUG >= 6
+	/* poison nmb so if a driver accesses uninitialized pointers, we
+	 * know what it is */
+	for (i = 0; i < sizeof(*nmb) / sizeof(unsigned long); i++)
+		((unsigned long *)nmb)[i] = i + 0x8a3fc1;
+#endif
+
 	nmb->wnd = wnd;
-	wnd->nmb->pdo = pdo;
+	nmb->pdo = pdo;
 	wd->wnd = wnd;
 	wnd->wd = wd;
 	nmb->filterdbs.eth_db = nmb;
@@ -1616,16 +1622,18 @@ static STDCALL NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	init_nmb_functions(nmb);
 	wnd->net_dev = net_dev;
 	wnd->ndis_irq = NULL;
-	nt_spin_lock_init(&wnd->xmit_lock);
+	init_MUTEX(&wnd->tx_ring_mutex);
+	nt_spin_lock_init(&wnd->tx_stats_lock);
 	init_MUTEX(&wnd->ndis_comm_mutex);
 	init_waitqueue_head(&wnd->ndis_comm_wq);
 	wnd->ndis_comm_done = 0;
 	/* don't send packets until the card is associated */
-	wnd->send_ok = 0;
-	INIT_WORK(&wnd->xmit_work, xmit_worker, wnd);
-	wnd->xmit_ring_start = 0;
-	wnd->xmit_ring_pending = 0;
-	nt_spin_lock_init(&wnd->send_packet_done_lock);
+	wnd->tx_ok = 0;
+	INIT_WORK(&wnd->tx_work, tx_worker, wnd);
+	wnd->tx_ring_start = 0;
+	wnd->tx_ring_end = 0;
+	wnd->is_tx_ring_full = 0;
+	nt_spin_lock_init(&wnd->tx_stats_lock);
 	wnd->encr_mode = Ndis802_11EncryptionDisabled;
 	wnd->auth_mode = Ndis802_11AuthModeOpen;
 	wnd->capa.encr = 0;
@@ -1646,12 +1654,7 @@ static STDCALL NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	set_bit(HW_AVAILABLE, &wnd->hw_status);
 	if (wd->driver->ndis_driver)
 		wd->driver->ndis_driver->miniport.shutdown = NULL;
-	/* ZyDas driver doesn't call completion function when
-	 * querying for stats or rssi, so disable stats */
-	if (stricmp(wd->driver->name, "zd1211u") == 0)
-		wnd->stats_enabled = FALSE;
-	else
-		wnd->stats_enabled = TRUE;
+	wnd->stats_enabled = TRUE;
 
 	fdo->reserved = wnd;
 	nmb = wnd->nmb;
