@@ -30,7 +30,7 @@ int debug = 0;
 
 /* use own workqueue instead of shared one, to avoid depriving
  * others */
-struct workqueue_struct *wrapper_wq;
+struct workqueue_struct *wrap_wq;
 
 WRAP_MODULE_PARM_STRING(if_name, 0400);
 MODULE_PARM_DESC(if_name, "Network interface name or template "
@@ -53,44 +53,13 @@ MODULE_PARM_DESC(hangcheck_interval, "The interval, in seconds, for checking"
 		 " if driver is hung. (default: 0)");
 
 MODULE_AUTHOR("ndiswrapper team <ndiswrapper-general@lists.sourceforge.net>");
+#ifdef MODULE_DESCRIPTION
+MODULE_DESCRIPTION("NDIS wrapper driver");
+#endif
 #ifdef MODULE_VERSION
 MODULE_VERSION(DRIVER_VERSION);
 #endif
-
-#ifdef USE_OWN_WORKQUEUE
-/* we need to get thread for the task running ndiswrapper_wq, so
- * schedule a worker for it soon after initializing ndiswrapper_wq */
-
-static struct work_struct _wrap_wq_init;
-static int _wrap_wq_init_state;
-#define WRAP_WQ_INIT 1
-#define WRAP_WQ_EXIT 2
-
-static void _wrap_wq_init_worker(void *data)
-{
-	struct task_struct *task;
-	struct nt_thread *thread;
-
-	task = current;
-	if (_wrap_wq_init_state == WRAP_WQ_INIT) {
-		thread = wrap_create_thread(task);
-		DBGTRACE1("task: %p, pid: %d, thread: %p",
-			  task, task->pid, thread);
-		if (!thread) {
-			_wrap_wq_init_state = -1;
-			return;
-		}
-	} else {
-		thread = KeGetCurrentThread();
-		if (thread) {
-			DBGTRACE1("task: %p, pid: %d, thread: %p",
-				  task, task->pid, thread);
-			wrap_remove_thread(thread);
-		}
-	}
-	_wrap_wq_init_state = 0;
-}
-#endif
+MODULE_LICENSE("GPL");
 
 static void module_cleanup(void)
 {
@@ -98,14 +67,10 @@ static void module_cleanup(void)
 #ifdef CONFIG_USB
 	usb_exit();
 #endif
+
 #ifdef USE_OWN_WORKQUEUE
-	_wrap_wq_init_state = WRAP_WQ_EXIT;
-	schedule_work(&_wrap_wq_init);
-	while (_wrap_wq_init_state) {
-		set_current_state(TASK_INTERRUPTIBLE);
-		schedule_timeout(4);
-	}
-	destroy_workqueue(wrapper_wq);
+	if (wrap_wq)
+		destroy_workqueue(wrap_wq);
 #endif
 	wrap_procfs_remove();
 	ndis_exit();
@@ -125,8 +90,13 @@ static int __init wrapper_init(void)
 	char *env[] = {NULL};
 	int ret;
 
-	printk(KERN_INFO "%s version %s loaded (preempt=%s,smp=%s)\n",
+	printk(KERN_INFO "%s version %s loaded (%spreempt=%s,smp=%s)\n",
 	       DRIVER_NAME, DRIVER_VERSION,
+#ifdef DEBUG
+	       UTS_RELEASE ",",
+#else
+	       "",
+#endif
 #if defined CONFIG_PREEMPT
 	       "yes",
 #else
@@ -139,24 +109,15 @@ static int __init wrapper_init(void)
 #endif
 		);
 
+#ifdef USE_OWN_WORKQUEUE
+	wrap_wq = create_singlethread_workqueue("wrap_wq");
+#endif
 	if (misc_funcs_init() || ntoskernel_init() || ndis_init()
 #ifdef CONFIG_USB
-	     || usb_init()
+	    || usb_init()
 #endif
 		)
 		goto err;
-#ifdef USE_OWN_WORKQUEUE
-	wrapper_wq = create_singlethread_workqueue("wrapper_wq");
-	INIT_WORK(&_wrap_wq_init, _wrap_wq_init_worker, 0);
-	_wrap_wq_init_state = WRAP_WQ_INIT;
-	schedule_work(&_wrap_wq_init);
-	while (_wrap_wq_init_state > 0) {
-		set_current_state(TASK_INTERRUPTIBLE);
-		schedule_timeout(4);
-	}
-	if (_wrap_wq_init_state < 0)
-		goto err;
-#endif
 	wrap_procfs_init();
 	if (loader_init())
 		goto err;
@@ -188,4 +149,3 @@ static void __exit wrapper_exit(void)
 module_init(wrapper_init);
 module_exit(wrapper_exit);
 
-MODULE_LICENSE("GPL");
